from warnings import warn

import numpy as np
from scipy import special
from scipy.special import gamma, genlaguerre

from dipy.core.geometry import cart2sphere
from dipy.core.gradients import gradient_table_from_gradient_strength_bvecs
from dipy.reconst import mapmri
from dipy.reconst.cache import Cache
from dipy.reconst.multi_voxel import multi_voxel_fit

try:  # preferred scipy >= 0.14, required scipy >= 1.0
    from scipy.special import factorial
except ImportError:
    from scipy.misc import factorial
import random

from scipy.optimize import fmin_l_bfgs_b

import dipy.reconst.dti as dti
from dipy.reconst.shm import real_sh_descoteaux_from_index
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")
plt, have_plt, _ = optional_package("matplotlib.pyplot")


class QtdmriModel(Cache):
    r"""The q$\tau$-dMRI model to analytically and continuously represent the
    q$\tau$ diffusion signal attenuation over diffusion sensitization q and
    diffusion time $\tau$.

    The model :footcite:p:`Fick2018` can be seen as an extension of the MAP-MRI
    basis :footcite:p:`Ozarslan2013` towards different diffusion times.

    The main idea is to model the diffusion signal over time and space as
    a linear combination of continuous functions,

    .. math::
        :nowrap:

        \hat{E}(\textbf{q},\tau;\textbf{c}) =
        \sum_i^{N_{\textbf{q}}}\sum_k^{N_\tau} \textbf{c}_{ik}
        \,\Phi_i(\textbf{q})\,T_k(\tau)

    where $\Phi$ and $T$ are the spatial and temporal basis functions,
    $N_{\textbf{q}}$ and $N_\tau$ are the maximum spatial and temporal
    order, and $i,k$ are basis order iterators.

    The estimation of the coefficients $c_i$ can be regularized using
    either analytic Laplacian regularization, sparsity regularization using
    the l1-norm, or both to do a type of elastic net regularization.

    From the coefficients, there exists an analytical formula to estimate
    the ODF, RTOP, RTAP, RTPP, QIV and MSD, for any diffusion time.

    Parameters
    ----------
    gtab : GradientTable,
        gradient directions and bvalues container class. The bvalues
        should be in the normal s/mm^2. big_delta and small_delta need to be
        given in seconds.
    radial_order : unsigned int,
        an even integer representing the spatial/radial order of the basis.
    time_order : unsigned int,
        an integer larger or equal than zero representing the time order
        of the basis.
    laplacian_regularization : bool,
        Regularize using the Laplacian of the qt-dMRI basis.
    laplacian_weighting: string or scalar,
        The string 'GCV' makes it use generalized cross-validation to find
        the regularization weight :footcite:p:`Craven1979`. A scalar sets the
        regularization weight to that value.
    l1_regularization : bool,
        Regularize by imposing sparsity in the coefficients using the
        l1-norm.
    l1_weighting : 'CV' or scalar,
        The string 'CV' makes it use five-fold cross-validation to find
        the regularization weight. A scalar sets the regularization weight
        to that value.
    cartesian : bool
        Whether to use the Cartesian or spherical implementation of the
        qt-dMRI basis, which we first explored in :footcite:p:`Fick2015`.
    anisotropic_scaling : bool
        Whether to use anisotropic scaling or isotropic scaling. This
        option can be used to test if the Cartesian implementation is
        equivalent with the spherical one when using the same scaling.
    normalization : bool
        Whether to normalize the basis functions such that their inner
        product is equal to one. Normalization is only necessary when
        imposing sparsity in the spherical basis if cartesian=False.
    constrain_q0 : bool
        whether to constrain the q0 point to unity along the tau-space.
        This is necessary to ensure that $E(0,\tau)=1$.
    bval_threshold : float
        the threshold b-value to be used, such that only data points below
        that threshold are used when estimating the scale factors.
    eigenvalue_threshold : float,
        Sets the minimum of the tensor eigenvalues in order to avoid
        stability problem.
    cvxpy_solver : str, optional
        cvxpy solver name. Optionally optimize the positivity constraint
        with a particular cvxpy solver. See https://www.cvxpy.org/ for
        details.
        Default: None (cvxpy chooses its own solver)

    References
    ----------
    .. footbibliography::
    """

    @warning_for_keywords()
    def __init__(
        self,
        gtab,
        *,
        radial_order=6,
        time_order=2,
        laplacian_regularization=False,
        laplacian_weighting=0.2,
        l1_regularization=False,
        l1_weighting=0.1,
        cartesian=True,
        anisotropic_scaling=True,
        normalization=False,
        constrain_q0=True,
        bval_threshold=1e10,
        eigenvalue_threshold=1e-04,
        cvxpy_solver="CLARABEL",
    ):
        if radial_order % 2 or radial_order < 0:
            msg = "radial_order must be zero or an even positive integer."
            msg += f" radial_order {radial_order} was given."
            raise ValueError(msg)

        if time_order < 0:
            msg = "time_order must be larger or equal than zero integer."
            msg += f" time_order {time_order} was given."
            raise ValueError(msg)

        if not isinstance(laplacian_regularization, bool):
            msg = "laplacian_regularization must be True or False."
            msg += f" Input value was {laplacian_regularization}."
            raise ValueError(msg)

        if laplacian_regularization:
            msg = "laplacian_regularization weighting must be 'GCV' "
            msg += "or a float larger or equal than zero."
            msg += f" Input value was {laplacian_weighting}."
            if isinstance(laplacian_weighting, str):
                if laplacian_weighting != "GCV":
                    raise ValueError(msg)
            elif isinstance(laplacian_weighting, float):
                if laplacian_weighting < 0:
                    raise ValueError(msg)
            else:
                raise ValueError(msg)

        if not isinstance(l1_regularization, bool):
            msg = "l1_regularization must be True or False."
            msg += f" Input value was {l1_regularization}."
            raise ValueError(msg)

        if l1_regularization:
            msg = "l1_weighting weighting must be 'CV' "
            msg += "or a float larger or equal than zero."
            msg += f" Input value was {l1_weighting}."
            if isinstance(l1_weighting, str):
                if l1_weighting != "CV":
                    raise ValueError(msg)
            elif isinstance(l1_weighting, float):
                if l1_weighting < 0:
                    raise ValueError(msg)
            else:
                raise ValueError(msg)

        if not isinstance(cartesian, bool):
            msg = "cartesian must be True or False."
            msg += f" Input value was {cartesian}."
            raise ValueError(msg)

        if not isinstance(anisotropic_scaling, bool):
            msg = "anisotropic_scaling must be True or False."
            msg += f" Input value was {anisotropic_scaling}."
            raise ValueError(msg)

        if not isinstance(constrain_q0, bool):
            msg = "constrain_q0 must be True or False."
            msg += f" Input value was {constrain_q0}."
            raise ValueError(msg)

        if not isinstance(bval_threshold, float) or bval_threshold < 0:
            msg = "bval_threshold must be a positive float."
            msg += f" Input value was {bval_threshold}."
            raise ValueError(msg)

        if not isinstance(eigenvalue_threshold, float) or eigenvalue_threshold < 0:
            msg = "eigenvalue_threshold must be a positive float."
            msg += f" Input value was {eigenvalue_threshold}."
            raise ValueError(msg)

        if laplacian_regularization or l1_regularization:
            if not have_cvxpy:
                msg = "cvxpy must be installed for Laplacian or l1 "
                msg += "regularization."
                raise ImportError(msg)
            if cvxpy_solver is not None:
                if cvxpy_solver not in cvxpy.installed_solvers():
                    msg = f"Input `cvxpy_solver` was set to {cvxpy_solver}."
                    msg += f" One of {', '.join(cvxpy.installed_solvers())}"
                    msg += " was expected."
                    raise ValueError(msg)

        if l1_regularization and not cartesian and not normalization:
            msg = "The non-Cartesian implementation must be normalized for the"
            msg += " l1-norm sparsity regularization to work. Set "
            msg += "normalization=True to proceed."
            raise ValueError(msg)

        self.gtab = gtab
        self.radial_order = radial_order
        self.time_order = time_order
        self.laplacian_regularization = laplacian_regularization
        self.laplacian_weighting = laplacian_weighting
        self.l1_regularization = l1_regularization
        self.l1_weighting = l1_weighting
        self.cartesian = cartesian
        self.anisotropic_scaling = anisotropic_scaling
        self.normalization = normalization
        self.constrain_q0 = constrain_q0
        self.bval_threshold = bval_threshold
        self.eigenvalue_threshold = eigenvalue_threshold
        self.cvxpy_solver = cvxpy_solver

        if self.cartesian:
            self.ind_mat = qtdmri_index_matrix(radial_order, time_order)
        else:
            self.ind_mat = qtdmri_isotropic_index_matrix(radial_order, time_order)

        # precompute parts of laplacian regularization matrices
        self.part4_reg_mat_tau = part4_reg_matrix_tau(self.ind_mat, 1.0)
        self.part23_reg_mat_tau = part23_reg_matrix_tau(self.ind_mat, 1.0)
        self.part1_reg_mat_tau = part1_reg_matrix_tau(self.ind_mat, 1.0)
        if self.cartesian:
            self.S_mat, self.T_mat, self.U_mat = mapmri.mapmri_STU_reg_matrices(
                radial_order
            )
        else:
            self.part1_uq_iso_precomp = (
                mapmri.mapmri_isotropic_laplacian_reg_matrix_from_index_matrix(
                    self.ind_mat[:, :3], 1.0
                )
            )

        self.tenmodel = dti.TensorModel(gtab)

    @multi_voxel_fit
    def fit(self, data, **kwargs):
        bval_mask = self.gtab.bvals < self.bval_threshold
        data_norm = data / data[self.gtab.b0s_mask].mean()
        tau = self.gtab.tau
        bvecs = self.gtab.bvecs
        qvals = self.gtab.qvals
        b0s_mask = self.gtab.b0s_mask

        if self.cartesian:
            if self.anisotropic_scaling:
                us, ut, R = qtdmri_anisotropic_scaling(
                    data_norm[bval_mask],
                    qvals[bval_mask],
                    bvecs[bval_mask],
                    tau[bval_mask],
                )
                tau_scaling = ut / us.mean()
                tau_scaled = tau * tau_scaling
                ut /= tau_scaling
                us = np.clip(us, self.eigenvalue_threshold, np.inf)
                q = np.dot(bvecs, R) * qvals[:, None]
                M = _qtdmri_signal_matrix(
                    self.radial_order,
                    self.time_order,
                    us,
                    ut,
                    q,
                    tau_scaled,
                    normalization=self.normalization,
                )
            else:
                us, ut = qtdmri_isotropic_scaling(data_norm, qvals, tau)
                tau_scaling = ut / us
                tau_scaled = tau * tau_scaling
                ut /= tau_scaling
                R = np.eye(3)
                us = np.tile(us, 3)
                q = bvecs * qvals[:, None]
                M = _qtdmri_signal_matrix(
                    self.radial_order,
                    self.time_order,
                    us,
                    ut,
                    q,
                    tau_scaled,
                    normalization=self.normalization,
                )
        else:
            us, ut = qtdmri_isotropic_scaling(data_norm, qvals, tau)
            tau_scaling = ut / us
            tau_scaled = tau * tau_scaling
            ut /= tau_scaling
            R = np.eye(3)
            us = np.tile(us, 3)
            q = bvecs * qvals[:, None]
            M = _qtdmri_isotropic_signal_matrix(
                self.radial_order,
                self.time_order,
                us[0],
                ut,
                q,
                tau_scaled,
                normalization=self.normalization,
            )

        b0_indices = np.arange(self.gtab.tau.shape[0])[self.gtab.b0s_mask]
        tau0_ordered = self.gtab.tau[b0_indices]
        unique_taus = np.unique(self.gtab.tau)
        first_tau_pos = []
        for unique_tau in unique_taus:
            first_tau_pos.append(np.where(tau0_ordered == unique_tau)[0][0])
        M0 = M[b0_indices[first_tau_pos]]

        lopt = 0.0
        alpha = 0.0
        if self.laplacian_regularization and not self.l1_regularization:
            if self.cartesian:
                laplacian_matrix = qtdmri_laplacian_reg_matrix(
                    self.ind_mat,
                    us,
                    ut,
                    S_mat=self.S_mat,
                    T_mat=self.T_mat,
                    U_mat=self.U_mat,
                    part1_ut_precomp=self.part1_reg_mat_tau,
                    part23_ut_precomp=self.part23_reg_mat_tau,
                    part4_ut_precomp=self.part4_reg_mat_tau,
                    normalization=self.normalization,
                )
            else:
                laplacian_matrix = qtdmri_isotropic_laplacian_reg_matrix(
                    self.ind_mat,
                    us,
                    ut,
                    part1_uq_iso_precomp=self.part1_uq_iso_precomp,
                    part1_ut_precomp=self.part1_reg_mat_tau,
                    part23_ut_precomp=self.part23_reg_mat_tau,
                    part4_ut_precomp=self.part4_reg_mat_tau,
                    normalization=self.normalization,
                )
            if self.laplacian_weighting == "GCV":
                try:
                    lopt = generalized_crossvalidation(data_norm, M, laplacian_matrix)
                except BaseException:
                    msg = "Laplacian GCV failed. lopt defaulted to 2e-4."
                    warn(msg, stacklevel=2)
                    lopt = 2e-4
            elif np.isscalar(self.laplacian_weighting):
                lopt = self.laplacian_weighting
            c = cvxpy.Variable(M.shape[1])
            design_matrix = cvxpy.Constant(M) @ c
            objective = cvxpy.Minimize(
                cvxpy.sum_squares(design_matrix - data_norm)
                + lopt * cvxpy.quad_form(c, laplacian_matrix)
            )
            if self.constrain_q0:
                # just constraint first and last, otherwise the solver fails
                constraints = [M0[0] @ c == 1, M0[-1] @ c == 1]
            else:
                constraints = []
            prob = cvxpy.Problem(objective, constraints)
            try:
                prob.solve(solver=self.cvxpy_solver, verbose=False)
                cvxpy_solution_optimal = prob.status == "optimal"
                qtdmri_coef = np.asarray(c.value).squeeze()
            except BaseException:
                qtdmri_coef = np.zeros(M.shape[1])
                cvxpy_solution_optimal = False
        elif self.l1_regularization and not self.laplacian_regularization:
            if self.l1_weighting == "CV":
                alpha = l1_crossvalidation(b0s_mask, data_norm, M)
            elif np.isscalar(self.l1_weighting):
                alpha = self.l1_weighting
            c = cvxpy.Variable(M.shape[1])
            design_matrix = cvxpy.Constant(M) @ c
            objective = cvxpy.Minimize(
                cvxpy.sum_squares(design_matrix - data_norm) + alpha * cvxpy.norm1(c)
            )
            if self.constrain_q0:
                # just constraint first and last, otherwise the solver fails
                constraints = [M0[0] @ c == 1, M0[-1] @ c == 1]
            else:
                constraints = []
            prob = cvxpy.Problem(objective, constraints)
            try:
                prob.solve(solver=self.cvxpy_solver, verbose=False)
                cvxpy_solution_optimal = prob.status == "optimal"
                qtdmri_coef = np.asarray(c.value).squeeze()
            except BaseException:
                qtdmri_coef = np.zeros(M.shape[1])
                cvxpy_solution_optimal = False
        elif self.l1_regularization and self.laplacian_regularization:
            if self.cartesian:
                laplacian_matrix = qtdmri_laplacian_reg_matrix(
                    self.ind_mat,
                    us,
                    ut,
                    S_mat=self.S_mat,
                    T_mat=self.T_mat,
                    U_mat=self.U_mat,
                    part1_ut_precomp=self.part1_reg_mat_tau,
                    part23_ut_precomp=self.part23_reg_mat_tau,
                    part4_ut_precomp=self.part4_reg_mat_tau,
                    normalization=self.normalization,
                )
            else:
                laplacian_matrix = qtdmri_isotropic_laplacian_reg_matrix(
                    self.ind_mat,
                    us,
                    ut,
                    part1_uq_iso_precomp=self.part1_uq_iso_precomp,
                    part1_ut_precomp=self.part1_reg_mat_tau,
                    part23_ut_precomp=self.part23_reg_mat_tau,
                    part4_ut_precomp=self.part4_reg_mat_tau,
                    normalization=self.normalization,
                )
            if self.laplacian_weighting == "GCV":
                lopt = generalized_crossvalidation(data_norm, M, laplacian_matrix)
            elif np.isscalar(self.laplacian_weighting):
                lopt = self.laplacian_weighting
            if self.l1_weighting == "CV":
                alpha = elastic_crossvalidation(
                    b0s_mask, data_norm, M, laplacian_matrix, lopt
                )
            elif np.isscalar(self.l1_weighting):
                alpha = self.l1_weighting
            c = cvxpy.Variable(M.shape[1])
            design_matrix = cvxpy.Constant(M) @ c
            objective = cvxpy.Minimize(
                cvxpy.sum_squares(design_matrix - data_norm)
                + alpha * cvxpy.norm1(c)
                + lopt * cvxpy.quad_form(c, laplacian_matrix)
            )
            if self.constrain_q0:
                # just constraint first and last, otherwise the solver fails
                constraints = [M0[0] @ c == 1, M0[-1] @ c == 1]
            else:
                constraints = []
            prob = cvxpy.Problem(objective, constraints)
            try:
                prob.solve(solver=self.cvxpy_solver, verbose=False)
                cvxpy_solution_optimal = prob.status == "optimal"
                qtdmri_coef = np.asarray(c.value).squeeze()
            except BaseException:
                qtdmri_coef = np.zeros(M.shape[1])
                cvxpy_solution_optimal = False
        elif not self.l1_regularization and not self.laplacian_regularization:
            # just use least squares with the observation matrix
            pseudoInv = np.linalg.pinv(M)
            qtdmri_coef = np.dot(pseudoInv, data_norm)
            # if cvxpy is used to constraint q0 without regularization the
            # solver often fails, so only first tau-position is manually
            # normalized.
            qtdmri_coef /= np.dot(M0[0], qtdmri_coef)
            cvxpy_solution_optimal = None

        if cvxpy_solution_optimal is False:
            msg = "cvxpy optimization resulted in non-optimal solution. Check "
            msg += "cvxpy_solution_optimal attribute in fitted object to see "
            msg += "which voxels are affected."
            warn(msg, stacklevel=2)
        return QtdmriFit(
            self,
            qtdmri_coef,
            us,
            ut,
            tau_scaling,
            R,
            lopt,
            alpha,
            cvxpy_solution_optimal,
        )


class QtdmriFit:
    def __init__(
        self,
        model,
        qtdmri_coef,
        us,
        ut,
        tau_scaling,
        R,
        lopt,
        alpha,
        cvxpy_solution_optimal,
    ):
        """Calculates diffusion properties for a single voxel

        Parameters
        ----------
        model : object,
            AnalyticalModel
        qtdmri_coef : 1d ndarray,
            qtdmri coefficients
        us : array, 3 x 1
            spatial scaling factors
        ut : float
            temporal scaling factor
        tau_scaling : float,
            the temporal scaling that used to scale tau to the size of us
        R : 3x3 numpy array,
            tensor eigenvectors
        lopt : float,
            laplacian regularization weight
        alpha : float,
            the l1 regularization weight
        cvxpy_solution_optimal: bool,
            indicates whether the cvxpy coefficient estimation reach an optimal
            solution
        """

        self.model = model
        self._qtdmri_coef = qtdmri_coef
        self.us = us
        self.ut = ut
        self.tau_scaling = tau_scaling
        self.R = R
        self.lopt = lopt
        self.alpha = alpha
        self.cvxpy_solution_optimal = cvxpy_solution_optimal

    def qtdmri_to_mapmri_coef(self, tau):
        """This function converts the qtdmri coefficients to mapmri
        coefficients for a given tau.

        Defined in :footcite:p:`Fick2018`, the conversion is performed by a
        matrix multiplication that evaluates the time-depenent part of the basis
        and multiplies it with the coefficients, after which coefficients with
        the same spatial orders are summed up, resulting in mapmri coefficients.

        Parameters
        ----------
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds

        References
        ----------
        .. footbibliography::
        """
        if self.model.cartesian:
            II = self.model.cache_get("qtdmri_to_mapmri_matrix", key=tau)
            if II is None:
                II = qtdmri_to_mapmri_matrix(
                    self.model.radial_order,
                    self.model.time_order,
                    self.ut,
                    self.tau_scaling * tau,
                )
                self.model.cache_set("qtdmri_to_mapmri_matrix", tau, II)
        else:
            II = self.model.cache_get("qtdmri_isotropic_to_mapmri_matrix", key=tau)
            if II is None:
                II = qtdmri_isotropic_to_mapmri_matrix(
                    self.model.radial_order,
                    self.model.time_order,
                    self.ut,
                    self.tau_scaling * tau,
                )
                self.model.cache_set("qtdmri_isotropic_to_mapmri_matrix", tau, II)
        mapmri_coef = np.dot(II, self._qtdmri_coef)
        return mapmri_coef

    @warning_for_keywords()
    def sparsity_abs(self, *, threshold=0.99):
        """As a measure of sparsity, calculates the number of largest
        coefficients needed to absolute sum up to 99% of the total absolute sum
        of all coefficients"""
        if not 0.0 < threshold < 1.0:
            msg = "sparsity threshold must be between zero and one"
            raise ValueError(msg)
        total_weight = np.sum(abs(self._qtdmri_coef))
        absolute_normalized_coef_array = (
            np.sort(abs(self._qtdmri_coef))[::-1] / total_weight
        )
        current_weight = 0.0
        counter = 0
        while current_weight < threshold:
            current_weight += absolute_normalized_coef_array[counter]
            counter += 1
        return counter

    @warning_for_keywords()
    def sparsity_density(self, *, threshold=0.99):
        """As a measure of sparsity, calculates the number of largest
        coefficients needed to squared sum up to 99% of the total squared sum
        of all coefficients"""
        if not 0.0 < threshold < 1.0:
            msg = "sparsity threshold must be between zero and one"
            raise ValueError(msg)
        total_weight = np.sum(self._qtdmri_coef**2)
        squared_normalized_coef_array = (
            np.sort(self._qtdmri_coef**2)[::-1] / total_weight
        )
        current_weight = 0.0
        counter = 0
        while current_weight < threshold:
            current_weight += squared_normalized_coef_array[counter]
            counter += 1
        return counter

    @warning_for_keywords()
    def odf(self, sphere, tau, *, s=2):
        r"""Calculates the analytical Orientation Distribution Function (ODF)
        for a given diffusion time tau from the signal.

        See :footcite:p:`Ozarslan2013` Eq. (32). The qtdmri coefficients are
        first converted to mapmri coefficients following :footcite:p:`Fick2018`.

        Parameters
        ----------
        sphere : dipy sphere object
            sphere object with vertice orientations to compute the ODF on.
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds
        s : unsigned int
            radial moment of the ODF

        References
        ----------
        .. footbibliography::
        """
        mapmri_coef = self.qtdmri_to_mapmri_coef(tau)
        if self.model.cartesian:
            v_ = sphere.vertices
            v = np.dot(v_, self.R)
            I_s = mapmri.mapmri_odf_matrix(self.model.radial_order, self.us, s, v)
            odf = np.dot(I_s, mapmri_coef)
        else:
            II = self.model.cache_get("ODF_matrix", key=(sphere, s))
            if II is None:
                II = mapmri.mapmri_isotropic_odf_matrix(
                    self.model.radial_order, 1, s, sphere.vertices
                )
                self.model.cache_set("ODF_matrix", (sphere, s), II)

            odf = self.us[0] ** s * np.dot(II, mapmri_coef)
        return odf

    @warning_for_keywords()
    def odf_sh(self, tau, *, s=2):
        r"""Calculates the real analytical odf for a given discrete sphere.

        Computes the design matrix of the ODF for the given sphere vertices
        and radial moment :footcite:p:`Ozarslan2013` eq. (32). The radial moment
        s acts as a sharpening method. The analytical equation for the spherical
        ODF basis is given in :footcite:p:`Fick2016b` eq. (C8). The qtdmri
        coefficients are first converted to mapmri coefficients following
        :footcite:p:`Fick2018`.

        Parameters
        ----------
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds
        s : unsigned int
            radial moment of the ODF

        References
        ----------
        .. footbibliography::
        """
        mapmri_coef = self.qtdmri_to_mapmri_coef(tau)
        if self.model.cartesian:
            msg = "odf in spherical harmonics not yet implemented for "
            msg += "cartesian implementation"
            raise ValueError(msg)
        II = self.model.cache_get("ODF_sh_matrix", key=(self.model.radial_order, s))

        if II is None:
            II = mapmri.mapmri_isotropic_odf_sh_matrix(self.model.radial_order, 1, s)
            self.model.cache_set("ODF_sh_matrix", (self.model.radial_order, s), II)

        odf = self.us[0] ** s * np.dot(II, mapmri_coef)
        return odf

    def rtpp(self, tau):
        r"""Calculates the analytical return to the plane probability (RTPP)
        for a given diffusion time tau.

        See :footcite:p:`Ozarslan2013` eq. (42). The analytical formula for the
        isotropic MAP-MRI basis was derived in :footcite:p:`Fick2016b` eq. (C11).
        The qtdmri coefficients are first converted to mapmri coefficients
        following :footcite:p:`Fick2018`.

        Parameters
        ----------
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds

        References
        ----------
        .. footbibliography::
        """
        mapmri_coef = self.qtdmri_to_mapmri_coef(tau)

        if self.model.cartesian:
            ind_mat = mapmri.mapmri_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat(ind_mat)
            sel = Bm > 0.0  # select only relevant coefficients
            const = 1 / (np.sqrt(2 * np.pi) * self.us[0])
            ind_sum = (-1.0) ** (ind_mat[sel, 0] / 2.0)
            rtpp_vec = const * Bm[sel] * ind_sum * mapmri_coef[sel]
            rtpp = rtpp_vec.sum()
            return rtpp
        else:
            ind_mat = mapmri.mapmri_isotropic_index_matrix(self.model.radial_order)
            rtpp_vec = np.zeros(int(ind_mat.shape[0]))
            count = 0
            for n in range(0, self.model.radial_order + 1, 2):
                for j in range(1, 2 + n // 2):
                    ll = n + 2 - 2 * j
                    const = (-1 / 2.0) ** (ll / 2) / np.sqrt(np.pi)
                    matsum = 0
                    for k in range(0, j):
                        matsum += (
                            (-1) ** k
                            * mapmri.binomialfloat(j + ll - 0.5, j - k - 1)
                            * gamma(ll / 2 + k + 1 / 2.0)
                            / (factorial(k) * 0.5 ** (ll / 2 + 1 / 2.0 + k))
                        )
                    for _ in range(-ll, ll + 1):
                        rtpp_vec[count] = const * matsum
                        count += 1
            direction = np.array(self.R[:, 0], ndmin=2)
            r, theta, phi = cart2sphere(
                direction[:, 0], direction[:, 1], direction[:, 2]
            )

            rtpp = (
                mapmri_coef
                * (1 / self.us[0])
                * rtpp_vec
                * real_sh_descoteaux_from_index(
                    ind_mat[:, 2], ind_mat[:, 1], theta, phi
                )
            )
            return rtpp.sum()

    def rtap(self, tau):
        r"""Calculates the analytical return to the axis probability (RTAP)
        for a given diffusion time tau.

        See :footcite:p:`Ozarslan2013` eq. (40, 44a). The analytical formula for
        the isotropic MAP-MRI basis was derived in :footcite:p:`Fick2016b` eq.
        (C11). The qtdmri coefficients are first converted to mapmri
        coefficients following :footcite:p:`Fick2018`.

        Parameters
        ----------
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds

        References
        ----------
        .. footbibliography::
        """
        mapmri_coef = self.qtdmri_to_mapmri_coef(tau)

        if self.model.cartesian:
            ind_mat = mapmri.mapmri_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat(ind_mat)
            sel = Bm > 0.0  # select only relevant coefficients
            const = 1 / (2 * np.pi * np.prod(self.us[1:]))
            ind_sum = (-1.0) ** (np.sum(ind_mat[sel, 1:], axis=1) / 2.0)
            rtap_vec = const * Bm[sel] * ind_sum * mapmri_coef[sel]
            rtap = np.sum(rtap_vec)
        else:
            ind_mat = mapmri.mapmri_isotropic_index_matrix(self.model.radial_order)
            rtap_vec = np.zeros(int(ind_mat.shape[0]))
            count = 0

            for n in range(0, self.model.radial_order + 1, 2):
                for j in range(1, 2 + n // 2):
                    ll = n + 2 - 2 * j
                    kappa = ((-1) ** (j - 1) * 2.0 ** (-(ll + 3) / 2.0)) / np.pi
                    matsum = 0
                    for k in range(0, j):
                        matsum += (
                            (-1) ** k
                            * mapmri.binomialfloat(j + ll - 0.5, j - k - 1)
                            * gamma((ll + 1) / 2.0 + k)
                        ) / (factorial(k) * 0.5 ** ((ll + 1) / 2.0 + k))
                    for _ in range(-ll, ll + 1):
                        rtap_vec[count] = kappa * matsum
                        count += 1
            rtap_vec *= 2

            direction = np.array(self.R[:, 0], ndmin=2)
            r, theta, phi = cart2sphere(
                direction[:, 0], direction[:, 1], direction[:, 2]
            )
            rtap_vec = (
                mapmri_coef
                * (1 / self.us[0] ** 2)
                * rtap_vec
                * real_sh_descoteaux_from_index(
                    ind_mat[:, 2], ind_mat[:, 1], theta, phi
                )
            )
            rtap = rtap_vec.sum()
        return rtap

    def rtop(self, tau):
        r"""Calculates the analytical return to the origin probability (RTOP)
        for a given diffusion time tau.

        See :footcite:p:`Ozarslan2013` eq. (36, 43). The analytical formula for
        the isotropic MAP-MRI basis was derived in :footcite:p:`Fick2016b` eq.
        (C11). The qtdmri coefficients are first converted to mapmri
        coefficients following :footcite:p:`Fick2018`.

        Parameters
        ----------
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds

        References
        ----------
        .. footbibliography::
        """
        mapmri_coef = self.qtdmri_to_mapmri_coef(tau)

        if self.model.cartesian:
            ind_mat = mapmri.mapmri_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat(ind_mat)
            const = 1 / (np.sqrt(8 * np.pi**3) * np.prod(self.us))
            ind_sum = (-1.0) ** (np.sum(ind_mat, axis=1) / 2)
            rtop_vec = const * ind_sum * Bm * mapmri_coef
            rtop = rtop_vec.sum()
        else:
            ind_mat = mapmri.mapmri_isotropic_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat_isotropic(ind_mat)
            const = 1 / (2 * np.sqrt(2.0) * np.pi ** (3 / 2.0))
            rtop_vec = const * (-1.0) ** (ind_mat[:, 0] - 1) * Bm
            rtop = (1 / self.us[0] ** 3) * rtop_vec * mapmri_coef
            rtop = rtop.sum()
        return rtop

    def msd(self, tau):
        r"""Calculates the analytical Mean Squared Displacement (MSD) for a
        given diffusion time tau.

        It is defined as the Laplacian of the origin of the estimated signal
        :footcite:t:`Cheng2012`. The analytical formula for the MAP-MRI basis
        was derived in :footcite:p:`Fick2016b` eq. (C13, D1). The qtdmri
        coefficients are first converted to mapmri coefficients following
        :footcite:p:`Fick2018`.

        Parameters
        ----------
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds

        References
        ----------
        .. footbibliography::
        """
        mapmri_coef = self.qtdmri_to_mapmri_coef(tau)
        mu = self.us
        if self.model.cartesian:
            ind_mat = mapmri.mapmri_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat(ind_mat)
            sel = Bm > 0.0  # select only relevant coefficients
            ind_sum = np.sum(ind_mat[sel], axis=1)
            nx, ny, nz = ind_mat[sel].T

            numerator = (
                (-1) ** (0.5 * (-ind_sum))
                * np.pi ** (3 / 2.0)
                * (
                    (1 + 2 * nx) * mu[0] ** 2
                    + (1 + 2 * ny) * mu[1] ** 2
                    + (1 + 2 * nz) * mu[2] ** 2
                )
            )

            denominator = (
                np.sqrt(
                    2.0 ** (-ind_sum) * factorial(nx) * factorial(ny) * factorial(nz)
                )
                * gamma(0.5 - 0.5 * nx)
                * gamma(0.5 - 0.5 * ny)
                * gamma(0.5 - 0.5 * nz)
            )

            msd_vec = mapmri_coef[sel] * (numerator / denominator)
            msd = msd_vec.sum()
        else:
            ind_mat = mapmri.mapmri_isotropic_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat_isotropic(ind_mat)
            sel = Bm > 0.0  # select only relevant coefficients
            msd_vec = (4 * ind_mat[sel, 0] - 1) * Bm[sel]
            msd = self.us[0] ** 2 * msd_vec * mapmri_coef[sel]
            msd = msd.sum()
        return msd

    def qiv(self, tau):
        r"""Calculates the analytical Q-space Inverse Variance (QIV) for given
        diffusion time tau.

        It is defined as the inverse of the Laplacian of the origin of the
        estimated propagator :footcite:p:`Hosseinbor2013` eq. (22). The
        analytical formula for the MAP-MRI basis was derived in
        :footcite:p:`Fick2016b` eq. (C14, D2). The qtdmri coefficients are first
        converted to mapmri coefficients following :footcite:t:`Fick2018`.

        Parameters
        ----------
        tau : float
            diffusion time (big_delta - small_delta / 3.) in seconds

        References
        ----------
        .. footbibliography::
        """
        mapmri_coef = self.qtdmri_to_mapmri_coef(tau)
        ux, uy, uz = self.us
        if self.model.cartesian:
            ind_mat = mapmri.mapmri_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat(ind_mat)
            sel = Bm > 0  # select only relevant coefficients
            nx, ny, nz = ind_mat[sel].T

            numerator = (
                8
                * np.pi**2
                * (ux * uy * uz) ** 3
                * np.sqrt(factorial(nx) * factorial(ny) * factorial(nz))
                * gamma(0.5 - 0.5 * nx)
                * gamma(0.5 - 0.5 * ny)
                * gamma(0.5 - 0.5 * nz)
            )

            denominator = np.sqrt(2.0 ** (-1 + nx + ny + nz)) * (
                (1 + 2 * nx) * uy**2 * uz**2
                + ux**2 * ((1 + 2 * nz) * uy**2 + (1 + 2 * ny) * uz**2)
            )

            qiv_vec = mapmri_coef[sel] * (numerator / denominator)
            qiv = qiv_vec.sum()
        else:
            ind_mat = mapmri.mapmri_isotropic_index_matrix(self.model.radial_order)
            Bm = mapmri.b_mat_isotropic(ind_mat)
            sel = Bm > 0.0  # select only relevant coefficients
            j = ind_mat[sel, 0]
            qiv_vec = (8 * (-1.0) ** (1 - j) * np.sqrt(2) * np.pi ** (7 / 2.0)) / (
                (4.0 * j - 1) * Bm[sel]
            )
            qiv = ux**5 * qiv_vec * mapmri_coef[sel]
            qiv = qiv.sum()
        return qiv

    @warning_for_keywords()
    def fitted_signal(self, *, gtab=None):
        """Recovers the fitted signal for the given gradient table. If no
        gradient table is given it recovers the signal for the gtab of the model
        object.
        """
        if gtab is None:
            E = self.predict(self.model.gtab)
        else:
            E = self.predict(gtab)
        return E

    @warning_for_keywords()
    def predict(self, qvals_or_gtab, *, S0=1.0):
        """Recovers the reconstructed signal for any qvalue array or gradient
        table.
        """
        tau_scaling = self.tau_scaling
        if isinstance(qvals_or_gtab, np.ndarray):
            q = qvals_or_gtab[:, :3]
            tau = qvals_or_gtab[:, 3] * tau_scaling
        else:
            gtab = qvals_or_gtab
            qvals = gtab.qvals
            tau = gtab.tau * tau_scaling
            q = qvals[:, None] * gtab.bvecs

        if self.model.cartesian:
            if self.model.anisotropic_scaling:
                q_rot = np.dot(q, self.R)
                M = _qtdmri_signal_matrix(
                    self.model.radial_order,
                    self.model.time_order,
                    self.us,
                    self.ut,
                    q_rot,
                    tau,
                    normalization=self.model.normalization,
                )
            else:
                M = _qtdmri_signal_matrix(
                    self.model.radial_order,
                    self.model.time_order,
                    self.us,
                    self.ut,
                    q,
                    tau,
                    normalization=self.model.normalization,
                )
        else:
            M = _qtdmri_isotropic_signal_matrix(
                self.model.radial_order,
                self.model.time_order,
                self.us[0],
                self.ut,
                q,
                tau,
                normalization=self.model.normalization,
            )
        E = S0 * np.dot(M, self._qtdmri_coef)
        return E

    def norm_of_laplacian_signal(self):
        """Calculates the norm of the laplacian of the fitted signal.

        This information could be useful to assess if the extrapolation of the
        fitted signal contains spurious oscillations. A high laplacian norm may
        indicate that these are present, and any q-space indices that
        use integrals of the signal may be corrupted (e.g. RTOP, RTAP, RTPP,
        QIV). In contrast to :footcite:t:`Fick2016b`, the Laplacian now
        describes oscillations in the 4-dimensional qt-signal
        :footcite:p:`Fick2018`.

        See :footcite:p:`Fick2016b` for a definition of the metric.

        References
        ----------
        .. footbibliography::
        """
        if self.model.cartesian:
            lap_matrix = qtdmri_laplacian_reg_matrix(
                self.model.ind_mat,
                self.us,
                self.ut,
                S_mat=self.model.S_mat,
                T_mat=self.model.T_mat,
                U_mat=self.model.U_mat,
                part1_ut_precomp=self.model.part1_reg_mat_tau,
                part23_ut_precomp=self.model.part23_reg_mat_tau,
                part4_ut_precomp=self.model.part4_reg_mat_tau,
                normalization=self.model.normalization,
            )
        else:
            lap_matrix = qtdmri_isotropic_laplacian_reg_matrix(
                self.model.ind_mat,
                self.us,
                self.ut,
                part1_uq_iso_precomp=self.model.part1_uq_iso_precomp,
                part1_ut_precomp=self.model.part1_reg_mat_tau,
                part23_ut_precomp=self.model.part23_reg_mat_tau,
                part4_ut_precomp=self.model.part4_reg_mat_tau,
                normalization=self.model.normalization,
            )
        norm_laplacian = np.dot(
            self._qtdmri_coef, np.dot(self._qtdmri_coef, lap_matrix)
        )
        return norm_laplacian

    def pdf(self, rt_points):
        """Diffusion propagator on a given set of real points.
        if the array r_points is non writeable, then intermediate
        results are cached for faster recalculation
        """
        tau_scaling = self.tau_scaling
        rt_points_ = rt_points * np.r_[1, 1, 1, tau_scaling]
        if self.model.cartesian:
            K = _qtdmri_eap_matrix(
                self.model.radial_order,
                self.model.time_order,
                self.us,
                self.ut,
                rt_points_,
                normalization=self.model.normalization,
            )
        else:
            K = _qtdmri_isotropic_eap_matrix(
                self.model.radial_order,
                self.model.time_order,
                self.us[0],
                self.ut,
                rt_points_,
                normalization=self.model.normalization,
            )
        eap = np.dot(K, self._qtdmri_coef)
        return eap


def _qtdmri_to_mapmri_matrix(radial_order, time_order, ut, tau, isotropic):
    """Generate the matrix that maps the spherical qtdmri coefficients to
    MAP-MRI coefficients.

    The conversion is done by only evaluating the time basis for a diffusion
    time tau and summing up coefficients with the same spatial basis orders
    :footcite:p:`Fick2018`.

    Parameters
    ----------
    radial_order : unsigned int,
        an even integer representing the spatial/radial order of the basis.
    time_order : unsigned int,
        an integer larger or equal than zero representing the time order
        of the basis.
    ut : float
        temporal scaling factor
    tau : float
        diffusion time (big_delta - small_delta / 3.) in seconds
    isotropic : bool
        `True` if the case is isotropic.

    References
    ----------
    .. footbibliography::
    """
    if isotropic:
        mapmri_ind_mat = mapmri.mapmri_isotropic_index_matrix(radial_order)
        qtdmri_ind_mat = qtdmri_isotropic_index_matrix(radial_order, time_order)
    else:
        mapmri_ind_mat = mapmri.mapmri_index_matrix(radial_order)
        qtdmri_ind_mat = qtdmri_index_matrix(radial_order, time_order)

    n_elem_mapmri = int(mapmri_ind_mat.shape[0])
    n_elem_qtdmri = int(qtdmri_ind_mat.shape[0])

    temporal_storage = np.zeros(time_order + 1)
    for o in range(time_order + 1):
        temporal_storage[o] = temporal_basis(o, ut, tau)

    counter = 0
    mapmri_mat = np.zeros((n_elem_mapmri, n_elem_qtdmri))
    for j, ll, m, o in qtdmri_ind_mat:
        index_overlap = np.all(
            [
                j == mapmri_ind_mat[:, 0],
                ll == mapmri_ind_mat[:, 1],
                m == mapmri_ind_mat[:, 2],
            ],
            0,
        )
        mapmri_mat[:, counter] = temporal_storage[o] * index_overlap
        counter += 1
    return mapmri_mat


def qtdmri_to_mapmri_matrix(radial_order, time_order, ut, tau):
    """Generate the matrix that maps the qtdmri coefficients to MAP-MRI
    coefficients for the anisotropic case.

    The conversion is done by only evaluating the time basis for a diffusion
    time tau and summing up coefficients with the same spatial basis orders
    :footcite:p:`Fick2018`.

    Parameters
    ----------
    radial_order : unsigned int,
        an even integer representing the spatial/radial order of the basis.
    time_order : unsigned int,
        an integer larger or equal than zero representing the time order
        of the basis.
    ut : float
        temporal scaling factor
    tau : float
        diffusion time (big_delta - small_delta / 3.) in seconds

    References
    ----------
    .. footbibliography::
    """

    return _qtdmri_to_mapmri_matrix(radial_order, time_order, ut, tau, False)


def qtdmri_isotropic_to_mapmri_matrix(radial_order, time_order, ut, tau):
    """Generate the matrix that maps the spherical qtdmri coefficients to
    MAP-MRI coefficients for the isotropic case.

    The conversion is done by only evaluating the time basis for a diffusion
    time tau and summing up coefficients with the same spatial basis orders
    :footcite:p:`Fick2018`.

    Parameters
    ----------
    radial_order : unsigned int,
        an even integer representing the spatial/radial order of the basis.
    time_order : unsigned int,
        an integer larger or equal than zero representing the time order
        of the basis.
    ut : float
        temporal scaling factor
    tau : float
        diffusion time (big_delta - small_delta / 3.) in seconds

    References
    ----------
    .. footbibliography::
    """

    return _qtdmri_to_mapmri_matrix(radial_order, time_order, ut, tau, True)


def qtdmri_temporal_normalization(ut):
    """Normalization factor for the temporal basis"""
    return np.sqrt(ut)


def qtdmri_mapmri_normalization(mu):
    """Normalization factor for Cartesian MAP-MRI basis. The scaling is the
    same for every basis function depending only on the spatial scaling
    mu.
    """
    sqrtC = np.sqrt(8 * np.prod(mu)) * np.pi ** (3.0 / 4.0)
    return sqrtC


def qtdmri_mapmri_isotropic_normalization(j, ell, u0):
    """Normalization factor for Spherical MAP-MRI basis. The normalization
    for a basis function with orders [j,l,m] depends only on orders j,l and
    the isotropic scale factor.
    """
    sqrtC = (2 * np.pi) ** (3.0 / 2.0) * np.sqrt(
        2**ell * u0**3 * gamma(j) / gamma(j + ell + 1.0 / 2.0)
    )
    return sqrtC


@warning_for_keywords()
def _qtdmri_signal_matrix(
    radial_order, time_order, us, ut, q, tau, *, normalization=False
):
    """Function to generate the qtdmri signal basis."""
    M = qtdmri_signal_matrix(radial_order, time_order, us, ut, q, tau)
    if normalization:
        sqrtC = qtdmri_mapmri_normalization(us)
        sqrtut = qtdmri_temporal_normalization(ut)
        sqrtCut = sqrtC * sqrtut
        M *= sqrtCut
    return M


def qtdmri_signal_matrix(radial_order, time_order, us, ut, q, tau):
    r"""Constructs the design matrix as a product of 3 separated radial,
    angular and temporal design matrices. It precomputes the relevant basis
    orders for each one and finally puts them together according to the index
    matrix
    """
    ind_mat = qtdmri_index_matrix(radial_order, time_order)

    n_dat = int(q.shape[0])
    n_elem = int(ind_mat.shape[0])
    qx, qy, qz = q.T
    mux, muy, muz = us

    temporal_storage = np.zeros((n_dat, time_order + 1))
    for o in range(time_order + 1):
        temporal_storage[:, o] = temporal_basis(o, ut, tau)

    Qx_storage = np.array(np.zeros((n_dat, radial_order + 1 + 4)), dtype=complex)
    Qy_storage = np.array(np.zeros((n_dat, radial_order + 1 + 4)), dtype=complex)
    Qz_storage = np.array(np.zeros((n_dat, radial_order + 1 + 4)), dtype=complex)
    for n in range(radial_order + 1 + 4):
        Qx_storage[:, n] = mapmri.mapmri_phi_1d(n, qx, mux)
        Qy_storage[:, n] = mapmri.mapmri_phi_1d(n, qy, muy)
        Qz_storage[:, n] = mapmri.mapmri_phi_1d(n, qz, muz)

    counter = 0
    Q = np.zeros((n_dat, n_elem))
    for nx, ny, nz, o in ind_mat:
        Q[:, counter] = (
            np.real(Qx_storage[:, nx] * Qy_storage[:, ny] * Qz_storage[:, nz])
            * temporal_storage[:, o]
        )
        counter += 1

    return Q


def qtdmri_eap_matrix(radial_order, time_order, us, ut, grid):
    r"""Constructs the design matrix as a product of 3 separated radial,
    angular and temporal design matrices. It precomputes the relevant basis
    orders for each one and finally puts them together according to the index
    matrix
    """
    ind_mat = qtdmri_index_matrix(radial_order, time_order)
    rx, ry, rz, tau = grid.T

    n_dat = int(rx.shape[0])
    n_elem = int(ind_mat.shape[0])
    mux, muy, muz = us

    temporal_storage = np.zeros((n_dat, time_order + 1))
    for o in range(time_order + 1):
        temporal_storage[:, o] = temporal_basis(o, ut, tau)

    Kx_storage = np.zeros((n_dat, radial_order + 1))
    Ky_storage = np.zeros((n_dat, radial_order + 1))
    Kz_storage = np.zeros((n_dat, radial_order + 1))
    for n in range(radial_order + 1):
        Kx_storage[:, n] = mapmri.mapmri_psi_1d(n, rx, mux)
        Ky_storage[:, n] = mapmri.mapmri_psi_1d(n, ry, muy)
        Kz_storage[:, n] = mapmri.mapmri_psi_1d(n, rz, muz)

    counter = 0
    K = np.zeros((n_dat, n_elem))
    for nx, ny, nz, o in ind_mat:
        K[:, counter] = (
            Kx_storage[:, nx]
            * Ky_storage[:, ny]
            * Kz_storage[:, nz]
            * temporal_storage[:, o]
        )
        counter += 1

    return K


@warning_for_keywords()
def _qtdmri_isotropic_signal_matrix(
    radial_order, time_order, us, ut, q, tau, *, normalization=False
):
    M = qtdmri_isotropic_signal_matrix(radial_order, time_order, us, ut, q, tau)
    if normalization:
        ind_mat = qtdmri_isotropic_index_matrix(radial_order, time_order)
        j, ll = ind_mat[:, :2].T
        sqrtut = qtdmri_temporal_normalization(ut)
        sqrtC = qtdmri_mapmri_isotropic_normalization(j, ll, us)
        sqrtCut = sqrtC * sqrtut
        M = M * sqrtCut[None, :]
    return M


def qtdmri_isotropic_signal_matrix(radial_order, time_order, us, ut, q, tau):
    ind_mat = qtdmri_isotropic_index_matrix(radial_order, time_order)
    qvals, theta, phi = cart2sphere(q[:, 0], q[:, 1], q[:, 2])

    n_dat = int(qvals.shape[0])
    n_elem = int(ind_mat.shape[0])

    num_j = int(np.max(ind_mat[:, 0]))
    num_o = int(time_order + 1)
    num_l = int(radial_order // 2 + 1)
    num_m = int(radial_order * 2 + 1)

    # Radial Basis
    radial_storage = np.zeros([num_j, num_l, n_dat])
    for j in range(1, num_j + 1):
        for ll in range(0, radial_order + 1, 2):
            radial_storage[j - 1, ll // 2, :] = radial_basis_opt(j, ll, us, qvals)

    # Angular Basis
    angular_storage = np.zeros([num_l, num_m, n_dat])
    for ll in range(0, radial_order + 1, 2):
        for m in range(-ll, ll + 1):
            angular_storage[ll // 2, m + ll, :] = angular_basis_opt(
                ll, m, qvals, theta, phi
            )

    # Temporal Basis
    temporal_storage = np.zeros([num_o + 1, n_dat])
    for o in range(0, num_o + 1):
        temporal_storage[o, :] = temporal_basis(o, ut, tau)

    # Construct full design matrix
    M = np.zeros((n_dat, n_elem))
    counter = 0
    for j, ll, m, o in ind_mat:
        M[:, counter] = (
            radial_storage[j - 1, ll // 2, :]
            * angular_storage[ll // 2, m + ll, :]
            * temporal_storage[o, :]
        )
        counter += 1
    return M


@warning_for_keywords()
def _qtdmri_eap_matrix(radial_order, time_order, us, ut, grid, *, normalization=False):
    sqrtCut = 1.0
    if normalization:
        sqrtC = qtdmri_mapmri_normalization(us)
        sqrtut = qtdmri_temporal_normalization(ut)
        sqrtCut = sqrtC * sqrtut
    K_tau = qtdmri_eap_matrix(radial_order, time_order, us, ut, grid) * sqrtCut
    return K_tau


@warning_for_keywords()
def _qtdmri_isotropic_eap_matrix(
    radial_order, time_order, us, ut, grid, *, normalization=False
):
    K = qtdmri_isotropic_eap_matrix(radial_order, time_order, us, ut, grid)
    if normalization:
        ind_mat = qtdmri_isotropic_index_matrix(radial_order, time_order)
        j, ll = ind_mat[:, :2].T
        sqrtut = qtdmri_temporal_normalization(ut)
        sqrtC = qtdmri_mapmri_isotropic_normalization(j, ll, us)
        sqrtCut = sqrtC * sqrtut
        K = K * sqrtCut[None, :]
    return K


def qtdmri_isotropic_eap_matrix(radial_order, time_order, us, ut, grid):
    r"""Constructs the design matrix as a product of 3 separated radial,
    angular and temporal design matrices. It precomputes the relevant basis
    orders for each one and finally puts them together according to the index
    matrix
    """

    rx, ry, rz, tau = grid.T
    R, theta, phi = cart2sphere(rx, ry, rz)
    theta[np.isnan(theta)] = 0

    ind_mat = qtdmri_isotropic_index_matrix(radial_order, time_order)
    n_dat = int(R.shape[0])
    n_elem = int(ind_mat.shape[0])

    num_j = int(np.max(ind_mat[:, 0]))
    num_o = int(time_order + 1)
    num_l = int(radial_order / 2 + 1)
    num_m = int(radial_order * 2 + 1)

    # Radial Basis
    radial_storage = np.zeros([num_j, num_l, n_dat])
    for j in range(1, num_j + 1):
        for ll in range(0, radial_order + 1, 2):
            radial_storage[j - 1, ll // 2, :] = radial_basis_EAP_opt(j, ll, us, R)

    # Angular Basis
    angular_storage = np.zeros([num_j, num_l, num_m, n_dat])
    for j in range(1, num_j + 1):
        for ll in range(0, radial_order + 1, 2):
            for m in range(-ll, ll + 1):
                angular_storage[j - 1, ll // 2, m + ll, :] = angular_basis_EAP_opt(
                    j, ll, m, R, theta, phi
                )

    # Temporal Basis
    temporal_storage = np.zeros([num_o + 1, n_dat])
    for o in range(0, num_o + 1):
        temporal_storage[o, :] = temporal_basis(o, ut, tau)

    # Construct full design matrix
    M = np.zeros((n_dat, n_elem))
    counter = 0
    for j, ll, m, o in ind_mat:
        M[:, counter] = (
            radial_storage[j - 1, ll // 2, :]
            * angular_storage[j - 1, ll // 2, m + ll, :]
            * temporal_storage[o, :]
        )
        counter += 1
    return M


def radial_basis_opt(j, ell, us, q):
    """Spatial basis dependent on spatial scaling factor us"""
    const = (
        us**ell
        * np.exp(-2 * np.pi**2 * us**2 * q**2)
        * genlaguerre(j - 1, ell + 0.5)(4 * np.pi**2 * us**2 * q**2)
    )
    return const


def angular_basis_opt(ell, m, q, theta, phi):
    """Angular basis independent of spatial scaling factor us. Though it
    includes q, it is independent of the data and can be precomputed.
    """
    const = (
        (-1) ** (ell / 2)
        * np.sqrt(4.0 * np.pi)
        * (2 * np.pi**2 * q**2) ** (ell / 2)
        * real_sh_descoteaux_from_index(m, ell, theta, phi)
    )
    return const


def radial_basis_EAP_opt(j, ell, us, r):
    radial_part = (
        (us**3) ** (-1)
        / (us**2) ** (ell / 2)
        * np.exp(-(r**2) / (2 * us**2))
        * genlaguerre(j - 1, ell + 0.5)(r**2 / us**2)
    )
    return radial_part


def angular_basis_EAP_opt(j, ell, m, r, theta, phi):
    angular_part = (
        (-1) ** (j - 1)
        * (np.sqrt(2) * np.pi) ** (-1)
        * (r**2 / 2) ** (ell / 2)
        * real_sh_descoteaux_from_index(m, ell, theta, phi)
    )
    return angular_part


def temporal_basis(o, ut, tau):
    """Temporal basis dependent on temporal scaling factor ut"""
    const = np.exp(-ut * tau / 2.0) * special.laguerre(o)(ut * tau)
    return const


def qtdmri_index_matrix(radial_order, time_order):
    """Computes the SHORE basis order indices according to [1]."""
    index_matrix = []
    for n in range(0, radial_order + 1, 2):
        for i in range(0, n + 1):
            for j in range(0, n - i + 1):
                for o in range(0, time_order + 1):
                    index_matrix.append([n - i - j, j, i, o])

    return np.array(index_matrix)


def qtdmri_isotropic_index_matrix(radial_order, time_order):
    """Computes the SHORE basis order indices according to [1]."""
    index_matrix = []
    for n in range(0, radial_order + 1, 2):
        for j in range(1, 2 + n // 2):
            ll = n + 2 - 2 * j
            for m in range(-ll, ll + 1):
                for o in range(0, time_order + 1):
                    index_matrix.append([j, ll, m, o])
    return np.array(index_matrix)


@warning_for_keywords()
def qtdmri_laplacian_reg_matrix(
    ind_mat,
    us,
    ut,
    *,
    S_mat=None,
    T_mat=None,
    U_mat=None,
    part1_ut_precomp=None,
    part23_ut_precomp=None,
    part4_ut_precomp=None,
    normalization=False,
):
    """Computes the cartesian qt-dMRI Laplacian regularization matrix.

    If given, uses precomputed matrices for temporal and spatial regularization
    matrices to speed up computation. Follows the formulation of Appendix B
    in :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    if S_mat is None or T_mat is None or U_mat is None:
        radial_order = ind_mat[:, :3].max()
        S_mat, T_mat, U_mat = mapmri.mapmri_STU_reg_matrices(radial_order)

    part1_us = mapmri.mapmri_laplacian_reg_matrix(
        ind_mat[:, :3], us, S_mat, T_mat, U_mat
    )
    part23_us = part23_reg_matrix_q(ind_mat, U_mat, T_mat, us)
    part4_us = part4_reg_matrix_q(ind_mat, U_mat, us)

    if part1_ut_precomp is None:
        part1_ut = part1_reg_matrix_tau(ind_mat, ut)
    else:
        part1_ut = part1_ut_precomp / ut
    if part23_ut_precomp is None:
        part23_ut = part23_reg_matrix_tau(ind_mat, ut)
    else:
        part23_ut = part23_ut_precomp * ut
    if part4_ut_precomp is None:
        part4_ut = part4_reg_matrix_tau(ind_mat, ut)
    else:
        part4_ut = part4_ut_precomp * ut**3

    regularization_matrix = (
        part1_us * part1_ut + part23_us * part23_ut + part4_us * part4_ut
    )

    if normalization:
        temporal_normalization = qtdmri_temporal_normalization(ut) ** 2
        spatial_normalization = qtdmri_mapmri_normalization(us) ** 2
        regularization_matrix *= temporal_normalization * spatial_normalization
    return regularization_matrix


@warning_for_keywords()
def qtdmri_isotropic_laplacian_reg_matrix(
    ind_mat,
    us,
    ut,
    *,
    part1_uq_iso_precomp=None,
    part1_ut_precomp=None,
    part23_ut_precomp=None,
    part4_ut_precomp=None,
    normalization=False,
):
    """Computes the spherical qt-dMRI Laplacian regularization matrix.

    If given, uses precomputed matrices for temporal and spatial regularization
    matrices to speed up computation. Follows the formulation of Appendix C
    in :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    if part1_uq_iso_precomp is None:
        part1_us = mapmri.mapmri_isotropic_laplacian_reg_matrix_from_index_matrix(
            ind_mat[:, :3], us[0]
        )
    else:
        part1_us = part1_uq_iso_precomp * us[0]

    if part1_ut_precomp is None:
        part1_ut = part1_reg_matrix_tau(ind_mat, ut)
    else:
        part1_ut = part1_ut_precomp / ut

    if part23_ut_precomp is None:
        part23_ut = part23_reg_matrix_tau(ind_mat, ut)
    else:
        part23_ut = part23_ut_precomp * ut

    if part4_ut_precomp is None:
        part4_ut = part4_reg_matrix_tau(ind_mat, ut)
    else:
        part4_ut = part4_ut_precomp * ut**3

    part23_us = part23_iso_reg_matrix_q(ind_mat, us[0])
    part4_us = part4_iso_reg_matrix_q(ind_mat, us[0])

    regularization_matrix = (
        part1_us * part1_ut + part23_us * part23_ut + part4_us * part4_ut
    )

    if normalization:
        temporal_normalization = qtdmri_temporal_normalization(ut) ** 2
        j, ll = ind_mat[:, :2].T
        pre_spatial_norm = qtdmri_mapmri_isotropic_normalization(j, ll, us[0])
        spatial_normalization = np.outer(pre_spatial_norm, pre_spatial_norm)
        regularization_matrix *= temporal_normalization * spatial_normalization
    return regularization_matrix


def part23_reg_matrix_q(ind_mat, U_mat, T_mat, us):
    """Partial cartesian spatial Laplacian regularization matrix.

    The implementation follows the second line of Eq. (B2) in
    :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    ux, uy, uz = us
    x, y, z, _ = ind_mat.T
    n_elem = int(ind_mat.shape[0])
    LR = np.zeros((n_elem, n_elem))
    for i in range(n_elem):
        for k in range(i, n_elem):
            val = 0
            if x[i] == x[k] and y[i] == y[k]:
                val += (
                    (uz / (ux * uy))
                    * U_mat[x[i], x[k]]
                    * U_mat[y[i], y[k]]
                    * T_mat[z[i], z[k]]
                )
            if x[i] == x[k] and z[i] == z[k]:
                val += (
                    (uy / (ux * uz))
                    * U_mat[x[i], x[k]]
                    * T_mat[y[i], y[k]]
                    * U_mat[z[i], z[k]]
                )
            if y[i] == y[k] and z[i] == z[k]:
                val += (
                    (ux / (uy * uz))
                    * T_mat[x[i], x[k]]
                    * U_mat[y[i], y[k]]
                    * U_mat[z[i], z[k]]
                )
            LR[i, k] = LR[k, i] = val
    return LR


def part23_iso_reg_matrix_q(ind_mat, us):
    """Partial spherical spatial Laplacian regularization matrix.

    The implementation follows the equation below Eq. (C4) in
    :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    n_elem = int(ind_mat.shape[0])

    LR = np.zeros((n_elem, n_elem))

    for i in range(n_elem):
        for k in range(i, n_elem):
            if ind_mat[i, 1] == ind_mat[k, 1] and ind_mat[i, 2] == ind_mat[k, 2]:
                ji = ind_mat[i, 0]
                jk = ind_mat[k, 0]
                ll = ind_mat[i, 1]
                if ji == (jk + 1):
                    LR[i, k] = LR[k, i] = (
                        2.0 ** (-ll) * -gamma(3 / 2.0 + jk + ll) / gamma(jk)
                    )
                elif ji == jk:
                    LR[i, k] = LR[k, i] = (
                        2.0 ** (-(ll + 1))
                        * (1 - 4 * ji - 2 * ll)
                        * gamma(1 / 2.0 + ji + ll)
                        / gamma(ji)
                    )
                elif ji == (jk - 1):
                    LR[i, k] = LR[k, i] = (
                        2.0 ** (-ll) * -gamma(3 / 2.0 + ji + ll) / gamma(ji)
                    )
    return LR / us


def part4_reg_matrix_q(ind_mat, U_mat, us):
    """Partial cartesian spatial Laplacian regularization matrix.

    The implementation follows equation Eq. (B2) in :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    ux, uy, uz = us
    x, y, z, _ = ind_mat.T
    n_elem = int(ind_mat.shape[0])
    LR = np.zeros((n_elem, n_elem))
    for i in range(n_elem):
        for k in range(i, n_elem):
            if x[i] == x[k] and y[i] == y[k] and z[i] == z[k]:
                LR[i, k] = LR[k, i] = (
                    (1.0 / (ux * uy * uz))
                    * U_mat[x[i], x[k]]
                    * U_mat[y[i], y[k]]
                    * U_mat[z[i], z[k]]
                )
    return LR


def part4_iso_reg_matrix_q(ind_mat, us):
    """Partial spherical spatial Laplacian regularization matrix.

    The implementation follows the equation below Eq. (C4) in
    :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    n_elem = int(ind_mat.shape[0])
    LR = np.zeros((n_elem, n_elem))
    for i in range(n_elem):
        for k in range(i, n_elem):
            if (
                ind_mat[i, 0] == ind_mat[k, 0]
                and ind_mat[i, 1] == ind_mat[k, 1]
                and ind_mat[i, 2] == ind_mat[k, 2]
            ):
                ji = ind_mat[i, 0]
                ll = ind_mat[i, 1]
                LR[i, k] = LR[k, i] = (
                    2.0 ** (-(ll + 2))
                    * gamma(1 / 2.0 + ji + ll)
                    / (np.pi**2 * gamma(ji))
                )

    return LR / us**3


def part1_reg_matrix_tau(ind_mat, ut):
    """Partial temporal Laplacian regularization matrix.

    The implementation follows Appendix B in :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    n_elem = int(ind_mat.shape[0])
    LD = np.zeros((n_elem, n_elem))
    for i in range(n_elem):
        for k in range(i, n_elem):
            oi = ind_mat[i, 3]
            ok = ind_mat[k, 3]
            if oi == ok:
                LD[i, k] = LD[k, i] = 1.0 / ut
    return LD


def part23_reg_matrix_tau(ind_mat, ut):
    """Partial temporal Laplacian regularization matrix.

    The implementation follows Appendix B in :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    n_elem = int(ind_mat.shape[0])
    LD = np.zeros((n_elem, n_elem))
    for i in range(n_elem):
        for k in range(i, n_elem):
            oi = ind_mat[i, 3]
            ok = ind_mat[k, 3]
            if oi == ok:
                LD[i, k] = LD[k, i] = 1 / 2.0
            else:
                LD[i, k] = LD[k, i] = np.abs(oi - ok)
    return ut * LD


def part4_reg_matrix_tau(ind_mat, ut):
    """Partial temporal Laplacian regularization matrix.

    The implementation follows Appendix B in :footcite:p:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    n_elem = int(ind_mat.shape[0])
    LD = np.zeros((n_elem, n_elem))

    for i in range(n_elem):
        for k in range(i, n_elem):
            oi = ind_mat[i, 3]
            ok = ind_mat[k, 3]

            sum1 = 0
            for p in range(1, min([ok, oi]) + 1 + 1):
                sum1 += (oi - p) * (ok - p) * H(min([oi, ok]) - p)

            sum2 = 0
            for p in range(0, min(ok - 2, oi - 1) + 1):
                sum2 += p

            sum3 = 0
            for p in range(0, min(ok - 1, oi - 2) + 1):
                sum3 += p

            LD[i, k] = LD[k, i] = (
                0.25 * np.abs(oi - ok)
                + (1 / 16.0) * mapmri.delta(oi, ok)
                + min([oi, ok])
                + sum1
                + H(oi - 1)
                * H(ok - 1)
                * (
                    oi
                    + ok
                    - 2
                    + sum2
                    + sum3
                    + H(abs(oi - ok) - 1) * (abs(oi - ok) - 1) * min([ok - 1, oi - 1])
                )
            )
    return LD * ut**3


def H(value):
    """Step function of H(x)=1 if x>=0 and zero otherwise. Used for the
    temporal laplacian matrix."""
    if value >= 0:
        return 1
    return 0


@warning_for_keywords()
def generalized_crossvalidation(data, M, LR, *, startpoint=5e-4):
    r"""Generalized Cross Validation Function.

    See :footcite:p:`Craven1979` for further details about the method.

    References
    ----------
    .. footbibliography::
    """
    startpoint = 1e-4
    MMt = np.dot(M.T, M)
    K = len(data)
    input_stuff = (data, M, MMt, K, LR)

    bounds = ((1e-5, 1),)
    res = fmin_l_bfgs_b(
        GCV_cost_function,
        startpoint,
        args=(input_stuff,),
        approx_grad=True,
        bounds=bounds,
        disp=False,
        pgtol=1e-10,
        factr=10.0,
    )
    return res[0][0]


def GCV_cost_function(weight, arguments):
    r"""Generalized Cross Validation Function that is iterated.

    See :footcite:p:`Craven1979` for further details about the method.

    References
    ----------
    .. footbibliography::
    """
    data, M, MMt, K, LR = arguments
    S = np.dot(np.dot(M, np.linalg.pinv(MMt + weight * LR)), M.T)
    trS = np.trace(S)
    normyytilde = np.linalg.norm(data - np.dot(S, data), 2)
    gcv_value = normyytilde / (K - trS)
    return gcv_value


def qtdmri_isotropic_scaling(data, q, tau):
    """Constructs design matrix for fitting an exponential to the
    diffusion time points.
    """
    dataclip = np.clip(data, 1e-05, 1.0)
    logE = -np.log(dataclip)
    logE_q = logE / (2 * np.pi**2)
    logE_tau = logE * 2

    B_q = np.array([q * q])
    inv_B_q = np.linalg.pinv(B_q)

    B_tau = np.array([tau])
    inv_B_tau = np.linalg.pinv(B_tau)

    us = np.sqrt(np.dot(logE_q, inv_B_q)).item()
    ut = np.dot(logE_tau, inv_B_tau).item()
    return us, ut


def qtdmri_anisotropic_scaling(data, q, bvecs, tau):
    """Constructs design matrix for fitting an exponential to the
    diffusion time points.
    """
    dataclip = np.clip(data, 1e-05, 10e10)
    logE = -np.log(dataclip)
    logE_q = logE / (2 * np.pi**2)
    logE_tau = logE * 2

    B_q = design_matrix_spatial(bvecs, q)
    inv_B_q = np.linalg.pinv(B_q)
    A = np.dot(inv_B_q, logE_q)

    evals, R = dti.decompose_tensor(dti.from_lower_triangular(A))
    us = np.sqrt(evals)

    B_tau = np.array([tau])
    inv_B_tau = np.linalg.pinv(B_tau)

    ut = np.dot(logE_tau, inv_B_tau).item()

    return us, ut, R


def design_matrix_spatial(bvecs, qvals):
    """Constructs design matrix for DTI weighted least squares or
    least squares fitting. (Basser et al., 1994a)

    Parameters
    ----------
    bvecs : array (N x 3)
        unit b-vectors of the acquisition.
    qvals : array (N,)
        corresponding q-values in 1/mm

    Returns
    -------
    design_matrix : array (g,7)
        Design matrix or B matrix assuming Gaussian distributed tensor model
        design_matrix[j, :] = (Bxx, Byy, Bzz, Bxy, Bxz, Byz, dummy)
    """
    B = np.zeros((bvecs.shape[0], 6))
    B[:, 0] = bvecs[:, 0] * bvecs[:, 0] * 1.0 * qvals**2  # Bxx
    B[:, 1] = bvecs[:, 0] * bvecs[:, 1] * 2.0 * qvals**2  # Bxy
    B[:, 2] = bvecs[:, 1] * bvecs[:, 1] * 1.0 * qvals**2  # Byy
    B[:, 3] = bvecs[:, 0] * bvecs[:, 2] * 2.0 * qvals**2  # Bxz
    B[:, 4] = bvecs[:, 1] * bvecs[:, 2] * 2.0 * qvals**2  # Byz
    B[:, 5] = bvecs[:, 2] * bvecs[:, 2] * 1.0 * qvals**2  # Bzz
    return B


def create_rt_space_grid(
    grid_size_r, max_radius_r, grid_size_tau, min_radius_tau, max_radius_tau
):
    """Generates EAP grid (for potential positivity constraint)."""
    tau_list = np.linspace(min_radius_tau, max_radius_tau, grid_size_tau)
    constraint_grid_tau = np.c_[0.0, 0.0, 0.0, 0.0]
    for tau in tau_list:
        constraint_grid = mapmri.create_rspace(grid_size_r, max_radius_r)
        constraint_grid_tau = np.vstack(
            [
                constraint_grid_tau,
                np.c_[constraint_grid, np.zeros(constraint_grid.shape[0]) + tau],
            ]
        )
    return constraint_grid_tau[1:]


def qtdmri_number_of_coefficients(radial_order, time_order):
    """Computes the total number of coefficients of the qtdmri basis given a
    radial and temporal order.

    See  equation given below Eq (9) in :footcite:t:`Fick2018`.

    References
    ----------
    .. footbibliography::
    """
    F = np.floor(radial_order / 2.0)
    Msym = (F + 1) * (F + 2) * (4 * F + 3) / 6
    M_total = Msym * (time_order + 1)
    return M_total


@warning_for_keywords()
def l1_crossvalidation(b0s_mask, E, M, *, weight_array=None):
    """cross-validation function to find the optimal weight of alpha for
    sparsity regularization"""

    if weight_array is None:
        weight_array = np.linspace(0, 0.4, 21)

    dwi_mask = ~b0s_mask
    b0_mask = b0s_mask
    dwi_indices = np.arange(E.shape[0])[dwi_mask]
    b0_indices = np.arange(E.shape[0])[b0_mask]
    random.shuffle(dwi_indices)

    sub0 = dwi_indices[0::5]
    sub1 = dwi_indices[1::5]
    sub2 = dwi_indices[2::5]
    sub3 = dwi_indices[3::5]
    sub4 = dwi_indices[4::5]

    test0 = np.hstack((b0_indices, sub1, sub2, sub3, sub4))
    test1 = np.hstack((b0_indices, sub0, sub2, sub3, sub4))
    test2 = np.hstack((b0_indices, sub0, sub1, sub3, sub4))
    test3 = np.hstack((b0_indices, sub0, sub1, sub2, sub4))
    test4 = np.hstack((b0_indices, sub0, sub1, sub2, sub3))

    cv_list = (
        (sub0, test0),
        (sub1, test1),
        (sub2, test2),
        (sub3, test3),
        (sub4, test4),
    )

    errorlist = np.zeros((5, 21))
    errorlist[:, 0] = 100.0
    optimal_alpha_sub = np.zeros(5)
    for i, (sub, test) in enumerate(cv_list):
        counter = 1
        cv_old = errorlist[i, 0]
        cv_new = errorlist[i, 0]
        while cv_old >= cv_new and counter < weight_array.shape[0]:
            alpha = weight_array[counter]
            c = cvxpy.Variable(M.shape[1])
            design_matrix = cvxpy.Constant(M[test]) @ c
            recovered_signal = cvxpy.Constant(M[sub]) @ c
            data = cvxpy.Constant(E[test])
            objective = cvxpy.Minimize(
                cvxpy.sum_squares(design_matrix - data) + alpha * cvxpy.norm1(c)
            )
            constraints = []
            prob = cvxpy.Problem(objective, constraints)
            prob.solve(solver="CLARABEL", verbose=False)
            errorlist[i, counter] = np.mean(
                (E[sub] - np.asarray(recovered_signal.value).squeeze()) ** 2
            )
            cv_old = errorlist[i, counter - 1]
            cv_new = errorlist[i, counter]
            counter += 1
        optimal_alpha_sub[i] = weight_array[counter - 1]
    optimal_alpha = optimal_alpha_sub.mean()
    return optimal_alpha


@warning_for_keywords()
def elastic_crossvalidation(b0s_mask, E, M, L, lopt, *, weight_array=None):
    """cross-validation function to find the optimal weight of alpha for
    sparsity regularization when also Laplacian regularization is used."""

    if weight_array is None:
        weight_array = np.linspace(0, 0.2, 21)

    dwi_mask = ~b0s_mask
    b0_mask = b0s_mask
    dwi_indices = np.arange(E.shape[0])[dwi_mask]
    b0_indices = np.arange(E.shape[0])[b0_mask]
    random.shuffle(dwi_indices)

    sub0 = dwi_indices[0::5]
    sub1 = dwi_indices[1::5]
    sub2 = dwi_indices[2::5]
    sub3 = dwi_indices[3::5]
    sub4 = dwi_indices[4::5]

    test0 = np.hstack((b0_indices, sub1, sub2, sub3, sub4))
    test1 = np.hstack((b0_indices, sub0, sub2, sub3, sub4))
    test2 = np.hstack((b0_indices, sub0, sub1, sub3, sub4))
    test3 = np.hstack((b0_indices, sub0, sub1, sub2, sub4))
    test4 = np.hstack((b0_indices, sub0, sub1, sub2, sub3))

    cv_list = (
        (sub0, test0),
        (sub1, test1),
        (sub2, test2),
        (sub3, test3),
        (sub4, test4),
    )

    errorlist = np.zeros((5, 21))
    errorlist[:, 0] = 100.0
    optimal_alpha_sub = np.zeros(5)
    for i, (sub, test) in enumerate(cv_list):
        counter = 1
        cv_old = errorlist[i, 0]
        cv_new = errorlist[i, 0]
        c = cvxpy.Variable(M.shape[1])
        design_matrix = cvxpy.Constant(M[test]) @ c
        recovered_signal = cvxpy.Constant(M[sub]) @ c

        data = cvxpy.Constant(E[test])
        constraints = []
        while cv_old >= cv_new and counter < weight_array.shape[0]:
            alpha = weight_array[counter]
            objective = cvxpy.Minimize(
                cvxpy.sum_squares(design_matrix - data)
                + alpha * cvxpy.norm1(c)
                + lopt * cvxpy.quad_form(c, L)
            )
            prob = cvxpy.Problem(objective, constraints)
            prob.solve(solver="CLARABEL", verbose=False)
            errorlist[i, counter] = np.mean(
                (E[sub] - np.asarray(recovered_signal.value).squeeze()) ** 2
            )
            cv_old = errorlist[i, counter - 1]
            cv_new = errorlist[i, counter]
            counter += 1
        optimal_alpha_sub[i] = weight_array[counter - 1]
    optimal_alpha = optimal_alpha_sub.mean()
    return optimal_alpha


@warning_for_keywords()
def visualise_gradient_table_G_Delta_rainbow(
    gtab,
    *,
    big_delta_start=None,
    big_delta_end=None,
    G_start=None,
    G_end=None,
    bval_isolines=np.r_[0, 250, 1000, 2500, 5000, 7500, 10000, 14000],
    alpha_shading=0.6,
):
    """This function visualizes a q-tau acquisition scheme as a function of
    gradient strength and pulse separation (big_delta). It represents every
    measurements at its G and big_delta position regardless of b-vector, with a
    background of b-value isolines for reference. It assumes there is only one
    unique pulse length (small_delta) in the acquisition scheme.

    Parameters
    ----------
    gtab : GradientTable object
        constructed gradient table with big_delta and small_delta given as
        inputs.
    big_delta_start : float,
        optional minimum big_delta that is plotted in seconds
    big_delta_end : float,
        optional maximum big_delta that is plotted in seconds
    G_start : float,
        optional minimum gradient strength that is plotted in T/m
    G_end : float,
        optional maximum gradient strength that is plotted in T/m
    bval_isolines : array,
        optional array of bvalue isolines that are plotted in the background
    alpha_shading : float between [0-1]
        optional shading of the bvalue colors in the background
    """
    Delta = gtab.big_delta  # in seconds
    delta = gtab.small_delta  # in seconds
    G = gtab.gradient_strength * 1e3  # in SI units T/m

    if len(np.unique(delta)) > 1:
        msg = "This acquisition has multiple small_delta values. "
        msg += "This visualization assumes there is only one small_delta."
        raise ValueError(msg)

    if big_delta_start is None:
        big_delta_start = 0.005
    if big_delta_end is None:
        big_delta_end = Delta.max() + 0.004
    if G_start is None:
        G_start = 0.0
    if G_end is None:
        G_end = G.max() + 0.05

    Delta_ = np.linspace(big_delta_start, big_delta_end, 50)
    G_ = np.linspace(G_start, G_end, 50)
    Delta_grid, G_grid = np.meshgrid(Delta_, G_)
    dummy_bvecs = np.tile([0, 0, 1], (len(G_grid.ravel()), 1))
    gtab_grid = gradient_table_from_gradient_strength_bvecs(
        G_grid.ravel() / 1e3, dummy_bvecs, Delta_grid.ravel(), delta[0]
    )
    bvals_ = gtab_grid.bvals.reshape(G_grid.shape)

    plt.contourf(
        Delta_, G_, bvals_, levels=bval_isolines, cmap="rainbow", alpha=alpha_shading
    )
    cb = plt.colorbar(spacing="proportional")
    cb.ax.tick_params(labelsize=16)
    plt.scatter(Delta, G, c="k", s=25)

    plt.xlim(big_delta_start, big_delta_end)
    plt.ylim(G_start, G_end)
    cb.set_label("b-value ($s$/$mm^2$)", fontsize=18)
    plt.xlabel(r"Pulse Separation $\Delta$ [sec]", fontsize=18)
    plt.ylabel("Gradient Strength [T/m]", fontsize=18)
    return None



"""
======================
MultiTensor Simulation
======================

In this example we show how someone can simulate the signal and the ODF of a
single voxel using a MultiTensor.
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.core.sphere import HemiSphere, disperse_charges
from dipy.data import get_sphere
from dipy.sims.voxel import multi_tensor, multi_tensor_odf
from dipy.viz import actor, window

###############################################################################
# For the simulation we will need a GradientTable with the b-values and
# b-vectors. To create one, we can first create some random points on a
# ``HemiSphere`` using spherical polar coordinates.

rng = np.random.default_rng()

n_pts = 64
theta = np.pi * rng.random(size=n_pts)
phi = 2 * np.pi * rng.random(size=n_pts)
hsph_initial = HemiSphere(theta=theta, phi=phi)

###############################################################################
# Next, we call ``disperse_charges`` which will iteratively move the points so
# that the electrostatic potential energy is minimized.

hsph_updated, potential = disperse_charges(hsph_initial, 5000)

###############################################################################
# We need two stacks of ``vertices``, one for every shell, and we need two sets
# of b-values, one at 1000 $s/mm^2$, and one at 2500 $s/mm^2$, as we discussed
# previously.

vertices = hsph_updated.vertices
values = np.ones(vertices.shape[0])

bvecs = np.vstack((vertices, vertices))
bvals = np.hstack((1000 * values, 2500 * values))

###############################################################################
# We can also add some b0s. Let's add one at the beginning and one at the end.

bvecs = np.insert(bvecs, (0, bvecs.shape[0]), np.array([0, 0, 0]), axis=0)
bvals = np.insert(bvals, (0, bvals.shape[0]), 0)

###############################################################################
# Let's now create the ``GradientTable``.

gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# In ``mevals`` we save the eigenvalues of each tensor.

mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])

###############################################################################
# In ``angles`` we save in polar coordinates (:math:`\theta, \phi`) the
# principal axis of each tensor.

angles = [(0, 0), (60, 0)]

###############################################################################
# In ``fractions`` we save the percentage of the contribution of each tensor.

fractions = [50, 50]

###############################################################################
# The function ``multi_tensor`` will return the simulated signal and an array
# with the principal axes of the tensors in cartesian coordinates.

signal, sticks = multi_tensor(
    gtab, mevals, S0=100, angles=angles, fractions=fractions, snr=None
)

###############################################################################
# We can also add Rician noise with a specific SNR.

signal_noisy, sticks = multi_tensor(
    gtab, mevals, S0=100, angles=angles, fractions=fractions, snr=20
)

plt.plot(signal, label="noiseless")

plt.plot(signal_noisy, label="with noise")
plt.legend()
# plt.show()
plt.savefig("simulated_signal.png", bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Simulated MultiTensor signal
#
#
#
# For the ODF simulation we will need a sphere. Because we are interested in a
# simulation of only a single voxel, we can use a sphere with very high
# resolution. We generate that by subdividing the triangles of one of DIPY_'s
# cached spheres, which we can read in the following way.

sphere = get_sphere(name="repulsion724")
sphere = sphere.subdivide(n=2)

odf = multi_tensor_odf(sphere.vertices, mevals, angles, fractions)

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()

odf_actor = actor.odf_slicer(odf[None, None, None, :], sphere=sphere, colormap="plasma")
odf_actor.RotateX(90)

scene.add(odf_actor)

print("Saving illustration as multi_tensor_simulation")
window.record(scene=scene, out_path="multi_tensor_simulation.png", size=(300, 300))
if interactive:
    window.show(scene)


###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Simulating a MultiTensor ODF.



import os
from tempfile import TemporaryDirectory

import pytest

from dipy.testing import assert_false, assert_true
from dipy.utils.optpkg import optional_package
from dipy.utils.tripwire import TripWireError


def test_optional_package():
    pkg, have_pkg, setup_module = optional_package("os")
    assert_true(have_pkg)
    assert_true(hasattr(pkg, "path"))

    pkg, have_pkg, setup_module = optional_package("not_a_package")
    assert_false(have_pkg)
    with pytest.raises(TripWireError):
        pkg.some_function()

    pkg, have_pkg, setup_module = optional_package("dipy", min_version="10.0.0")
    assert_false(have_pkg)
    with pytest.raises(TripWireError):
        pkg.some_function()

    pkg, have_pkg, setup_module = optional_package("dipy", min_version="1.0.0")
    assert_true(have_pkg)

    with TemporaryDirectory() as tmpdir:
        os.makedirs(os.path.join(tmpdir, "fake_package"))
        open(os.path.join(tmpdir, "fake_package", "__init__.py"), "a").close()
        current_dir = os.getcwd()
        os.chdir(tmpdir)
        pkg, have_pkg, setup_module = optional_package(
            "fake_package", min_version="1.0.0"
        )
        assert_false(have_pkg)
        assert_false(hasattr(pkg, "__version__"))
        with pytest.raises(TripWireError):
            pkg.some_function()
        os.chdir(current_dir)



from warnings import warn

import numpy as np
from scipy.optimize import leastsq
from scipy.special import gamma, hyp1f1

from dipy.core.geometry import cart2sphere
from dipy.data import default_sphere
from dipy.reconst.cache import Cache
from dipy.reconst.csdeconv import csdeconv
from dipy.reconst.multi_voxel import multi_voxel_fit
from dipy.reconst.odf import OdfFit, OdfModel
from dipy.reconst.shm import real_sh_descoteaux_from_index
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.deprecator import deprecated_params
from dipy.utils.optpkg import optional_package

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")


class ForecastModel(OdfModel, Cache):
    r"""Fiber ORientation Estimated using Continuous Axially Symmetric Tensors
    (FORECAST).

    FORECAST :footcite:p:`Anderson2005`, :footcite:p:`Kaden2016a`,
    :footcite:p:`Zucchelli2017` is a Spherical Deconvolution reconstruction
    model for multi-shell diffusion data which enables the calculation of a
    voxel adaptive response function using the Spherical Mean Technique (SMT)
    :footcite:p:`Kaden2016a`, :footcite:p:`Zucchelli2017`.

    With FORECAST it is possible to calculate crossing invariant parallel
    diffusivity, perpendicular diffusivity, mean diffusivity, and fractional
    anisotropy :footcite:p:`Kaden2016a`.

    References
    ----------
    .. footbibliography::

    Notes
    -----
    The implementation of FORECAST may require CVXPY (https://www.cvxpy.org/).
    """

    @deprecated_params("sh_order", new_name="sh_order_max", since="1.9", until="2.0")
    @warning_for_keywords()
    def __init__(
        self,
        gtab,
        *,
        sh_order_max=8,
        lambda_lb=1e-3,
        dec_alg="CSD",
        sphere=None,
        lambda_csd=1.0,
    ):
        r"""Analytical and continuous modeling of the diffusion signal with
        respect to the FORECAST basis.

        This implementation is a modification of the original FORECAST
        model presented in :footcite:p:`Anderson2005` adapted for multi-shell
        data as in :footcite:p:`Kaden2016a`, :footcite:p:`Zucchelli2017`.

        The main idea is to model the diffusion signal as the combination of a
        single fiber response function $F(\mathbf{b})$ times the fODF
        $\rho(\mathbf{v})$

        .. math::

            E(\mathbf{b}) = \int_{\mathbf{v} \in \mathcal{S}^2} \rho(\mathbf{v}) F({\mathbf{b}} | \mathbf{v}) d \mathbf{v}

        where $\mathbf{b}$ is the b-vector (b-value times gradient direction)
        and $\mathbf{v}$ is a unit vector representing a fiber direction.

        In FORECAST $\rho$ is modeled using real symmetric Spherical Harmonics
        (SH) and $F(\mathbf(b))$ is an axially symmetric tensor.


        Parameters
        ----------
        gtab : GradientTable,
            gradient directions and bvalues container class.
        sh_order_max : unsigned int, optional
            an even integer that represent the maximal SH order ($l$) of the
            basis (max 12)
        lambda_lb: float, optional
            Laplace-Beltrami regularization weight.
        dec_alg : str, optional
            Spherical deconvolution algorithm. The possible values are Weighted Least Squares ('WLS'),
            Positivity Constraints using CVXPY ('POS') and the Constraint
            Spherical Deconvolution algorithm ('CSD'). Default is 'CSD'.
        sphere : array, shape (N,3), optional
            sphere points where to enforce positivity when 'POS' or 'CSD'
            dec_alg are selected.
        lambda_csd : float, optional
            CSD regularization weight.

        References
        ----------
        .. footbibliography::

        Examples
        --------
        In this example, where the data, gradient table and sphere tessellation
        used for reconstruction are provided, we model the diffusion signal
        with respect to the FORECAST and compute the fODF, parallel and
        perpendicular diffusivity.

        >>> import warnings
        >>> from dipy.data import default_sphere, get_3shell_gtab
        >>> gtab = get_3shell_gtab()
        >>> from dipy.sims.voxel import multi_tensor
        >>> mevals = np.array(([0.0017, 0.0003, 0.0003],
        ...                    [0.0017, 0.0003, 0.0003]))
        >>> angl = [(0, 0), (60, 0)]
        >>> data, sticks = multi_tensor(gtab,
        ...                             mevals,
        ...                             S0=100.0,
        ...                             angles=angl,
        ...                             fractions=[50, 50],
        ...                             snr=None)
        >>> from dipy.reconst.forecast import ForecastModel
        >>> from dipy.reconst.shm import descoteaux07_legacy_msg
        >>> with warnings.catch_warnings():
        ...     warnings.filterwarnings(
        ...         "ignore", message=descoteaux07_legacy_msg,
        ...         category=PendingDeprecationWarning)
        ...     fm = ForecastModel(gtab, sh_order_max=6)
        >>> f_fit = fm.fit(data)
        >>> d_par = f_fit.dpar
        >>> d_perp = f_fit.dperp
        >>> with warnings.catch_warnings():
        ...     warnings.filterwarnings(
        ...         "ignore", message=descoteaux07_legacy_msg,
        ...         category=PendingDeprecationWarning)
        ...     fodf = f_fit.odf(default_sphere)
        """  # noqa: E501
        OdfModel.__init__(self, gtab)

        # round the bvals in order to avoid numerical errors
        self.bvals = np.round(gtab.bvals / 100) * 100
        self.bvecs = gtab.bvecs

        if 0 <= sh_order_max <= 12 and not bool(sh_order_max % 2):
            self.sh_order_max = sh_order_max
        else:
            msg = "sh_order_max must be a non-zero even positive number "
            msg += "between 2 and 12"
            raise ValueError(msg)

        if sphere is None:
            sphere = default_sphere
            self.vertices = sphere.vertices[0 : int(sphere.vertices.shape[0] / 2), :]

        else:
            self.vertices = sphere

        self.b0s_mask = self.bvals == 0
        self.one_0_bvals = np.r_[0, self.bvals[~self.b0s_mask]]
        self.one_0_bvecs = np.r_[
            np.array([0, 0, 0]).reshape(1, 3), self.bvecs[~self.b0s_mask, :]
        ]

        self.rho = rho_matrix(self.sh_order_max, self.one_0_bvecs)

        # signal regularization matrix
        self.srm = rho_matrix(4, self.one_0_bvecs)
        self.lb_matrix_signal = lb_forecast(4)

        self.b_unique = np.sort(np.unique(self.bvals[self.bvals > 0]))
        self.wls = True
        self.csd = False
        self.pos = False

        if dec_alg.upper() == "POS":
            if not have_cvxpy:
                cvxpy.import_error()

            self.wls = False
            self.pos = True

        if dec_alg.upper() == "CSD":
            self.csd = True

        self.lb_matrix = lb_forecast(self.sh_order_max)
        self.lambda_lb = lambda_lb
        self.lambda_csd = lambda_csd
        self.fod = rho_matrix(sh_order_max, self.vertices)

    @multi_voxel_fit
    def fit(self, data, **kwargs):
        data_b0 = data[self.b0s_mask].mean()
        data_single_b0 = np.r_[data_b0, data[~self.b0s_mask]]
        data_single_b0 = np.divide(
            data_single_b0,
            data_b0,
            out=np.zeros_like(data_single_b0),
            where=data_b0 != 0,
        )

        # calculates the mean signal at each b_values
        means = find_signal_means(
            self.b_unique,
            data_single_b0,
            self.one_0_bvals,
            self.srm,
            self.lb_matrix_signal,
        )

        # average diffusivity initialization
        x = np.array([np.pi / 4, np.pi / 4])

        x, status = leastsq(forecast_error_func, x, args=(self.b_unique, means))

        # transform to bound the diffusivities from 0 to 3e-03
        d_par = np.cos(x[0]) ** 2 * 3e-03
        d_perp = np.cos(x[1]) ** 2 * 3e-03

        if d_perp >= d_par:
            d_par, d_perp = d_perp, d_par

        # round to avoid memory explosion
        diff_key = str(int(np.round(d_par * 1e05))) + str(int(np.round(d_perp * 1e05)))

        M_diff = self.cache_get("forecast_matrix", key=diff_key)
        if M_diff is None:
            M_diff = forecast_matrix(self.sh_order_max, d_par, d_perp, self.one_0_bvals)
            self.cache_set("forecast_matrix", key=diff_key, value=M_diff)

        M = M_diff * self.rho
        M0 = M[:, 0]
        c0 = np.sqrt(1.0 / (4 * np.pi))

        # coefficients vector initialization
        n_c = int((self.sh_order_max + 1) * (self.sh_order_max + 2) / 2)
        coef = np.zeros(n_c)
        coef[0] = c0
        if int(np.round(d_par * 1e05)) > int(np.round(d_perp * 1e05)):
            if self.wls:
                data_r = data_single_b0 - M0 * c0

                Mr = M[:, 1:]
                Lr = self.lb_matrix[1:, 1:]

                pseudo_inv = np.dot(
                    np.linalg.inv(np.dot(Mr.T, Mr) + self.lambda_lb * Lr), Mr.T
                )
                coef = np.dot(pseudo_inv, data_r)
                coef = np.r_[c0, coef]

            if self.csd:
                coef, _ = csdeconv(data_single_b0, M, self.fod, tau=0.1, convergence=50)
                coef = coef / coef[0] * c0

            if self.pos:
                c = cvxpy.Variable(M.shape[1])
                design_matrix = cvxpy.Constant(M) @ c
                objective = cvxpy.Minimize(
                    cvxpy.sum_squares(design_matrix - data_single_b0)
                    + self.lambda_lb * cvxpy.quad_form(c, self.lb_matrix)
                )

                constraints = [c[0] == c0, self.fod @ c >= 0]
                prob = cvxpy.Problem(objective, constraints)
                try:
                    prob.solve(solver=cvxpy.OSQP, eps_abs=1e-05, eps_rel=1e-05)
                    coef = np.asarray(c.value).squeeze()
                except Exception:
                    warn("Optimization did not find a solution", stacklevel=2)
                    coef = np.zeros(M.shape[1])
                    coef[0] = c0

        return ForecastFit(self, data, coef, d_par, d_perp)


class ForecastFit(OdfFit):
    def __init__(self, model, data, sh_coef, d_par, d_perp):
        """Calculates diffusion properties for a single voxel

        Parameters
        ----------
        model : object,
            AnalyticalModel
        data : 1d ndarray,
            fitted data
        sh_coef : 1d ndarray,
            forecast sh coefficients
        d_par : float,
            parallel diffusivity
        d_perp : float,
            perpendicular diffusivity
        """
        OdfFit.__init__(self, model, data)
        self.model = model
        self._sh_coef = sh_coef
        self.gtab = model.gtab
        self.sh_order_max = model.sh_order_max
        self.d_par = d_par
        self.d_perp = d_perp

        self.rho = None

    @warning_for_keywords()
    def odf(self, sphere, *, clip_negative=True):
        r"""Calculates the fODF for a given discrete sphere.

        Parameters
        ----------
        sphere : Sphere,
            the odf sphere
        clip_negative : boolean, optional
            if True clip the negative odf values to 0, default True
        """
        if self.rho is None:
            self.rho = rho_matrix(self.sh_order_max, sphere.vertices)

        odf = np.dot(self.rho, self._sh_coef)

        if clip_negative:
            odf = np.clip(odf, 0, odf.max())

        return odf

    def fractional_anisotropy(self):
        r"""Calculates the fractional anisotropy."""
        fa = np.sqrt(
            0.5
            * (2 * (self.d_par - self.d_perp) ** 2)
            / (self.d_par**2 + 2 * self.d_perp**2)
        )
        return fa

    def mean_diffusivity(self):
        r"""Calculates the mean diffusivity."""
        md = (self.d_par + 2 * self.d_perp) / 3.0
        return md

    @warning_for_keywords()
    def predict(self, *, gtab=None, S0=1.0):
        r"""Calculates the fODF for a given discrete sphere.

        Parameters
        ----------
        gtab : GradientTable, optional
            gradient directions and bvalues container class.
        S0 : float, optional
            the signal at b-value=0

        """
        if gtab is None:
            gtab = self.gtab

        M_diff = forecast_matrix(self.sh_order_max, self.d_par, self.d_perp, gtab.bvals)

        rho = rho_matrix(self.sh_order_max, gtab.bvecs)
        M = M_diff * rho
        S = S0 * np.dot(M, self._sh_coef)

        return S

    @property
    def sh_coeff(self):
        """The FORECAST SH coefficients"""
        return self._sh_coef

    @property
    def dpar(self):
        """The parallel diffusivity"""
        return self.d_par

    @property
    def dperp(self):
        """The perpendicular diffusivity"""
        return self.d_perp


@warning_for_keywords()
def find_signal_means(b_unique, data_norm, bvals, rho, lb_matrix, *, w=1e-03):
    r"""Calculate the mean signal for each shell.

    Parameters
    ----------
    b_unique : 1d ndarray,
        unique b-values in a vector excluding zero
    data_norm : 1d ndarray,
        normalized diffusion signal
    bvals : 1d ndarray,
        the b-values
    rho : 2d ndarray,
        SH basis matrix for fitting the signal on each shell
    lb_matrix : 2d ndarray,
        Laplace-Beltrami regularization matrix
    w : float,
        weight for the Laplace-Beltrami regularization

    Returns
    -------
    means : 1d ndarray
        the average of the signal for each b-values

    """
    lb = len(b_unique)
    means = np.zeros(lb)
    for u in range(lb):
        ind = bvals == b_unique[u]
        shell = data_norm[ind]
        if np.sum(ind) > 20:
            M = rho[ind, :]
            coef = np.linalg.multi_dot(
                [np.linalg.inv(np.dot(M.T, M) + w * lb_matrix), M.T, shell]
            )

            means[u] = coef[0] / np.sqrt(4 * np.pi)
        else:
            means[u] = shell.mean()

    return means


def forecast_error_func(x, b_unique, E):
    r"""Calculates the difference between the mean signal calculated using
    the parameter vector x and the average signal E using FORECAST and SMT
    """
    d_par = np.cos(x[0]) ** 2 * 3e-03
    d_perp = np.cos(x[1]) ** 2 * 3e-03

    if d_perp >= d_par:
        d_par, d_perp = d_perp, d_par

    E_reconst = (
        0.5 * np.exp(-b_unique * d_perp) * psi_l(0, (b_unique * (d_par - d_perp)))
    )

    v = E - E_reconst
    return v


def psi_l(ell, b):
    n = ell // 2
    v = (-b) ** n
    v *= gamma(n + 1.0 / 2) / gamma(2 * n + 3.0 / 2)
    v *= hyp1f1(n + 1.0 / 2, 2 * n + 3.0 / 2, -b)
    return v


@deprecated_params("sh_order", new_name="sh_order_max", since="1.9", until="2.0")
def forecast_matrix(sh_order_max, d_par, d_perp, bvals):
    r"""Compute the FORECAST radial matrix"""
    n_c = int((sh_order_max + 1) * (sh_order_max + 2) / 2)
    M = np.zeros((bvals.shape[0], n_c))
    counter = 0
    for ell in range(0, sh_order_max + 1, 2):
        for _ in range(-ell, ell + 1):
            M[:, counter] = (
                2
                * np.pi
                * np.exp(-bvals * d_perp)
                * psi_l(ell, bvals * (d_par - d_perp))
            )
            counter += 1
    return M


@deprecated_params("sh_order", new_name="sh_order_max", since="1.9", until="2.0")
def rho_matrix(sh_order_max, vecs):
    r"""Compute the SH matrix $\rho$"""

    r, theta, phi = cart2sphere(vecs[:, 0], vecs[:, 1], vecs[:, 2])
    theta[np.isnan(theta)] = 0

    n_c = int((sh_order_max + 1) * (sh_order_max + 2) / 2)
    rho = np.zeros((vecs.shape[0], n_c))
    counter = 0
    for l_values in range(0, sh_order_max + 1, 2):
        for m_values in range(-l_values, l_values + 1):
            rho[:, counter] = real_sh_descoteaux_from_index(
                m_values, l_values, theta, phi
            )
            counter += 1
    return rho


@deprecated_params("sh_order", new_name="sh_order_max", since="1.9", until="2.0")
def lb_forecast(sh_order_max):
    r"""Returns the Laplace-Beltrami regularization matrix for FORECAST"""
    n_c = int((sh_order_max + 1) * (sh_order_max + 2) / 2)
    diag_lb = np.zeros(n_c)
    counter = 0
    for j in range(0, sh_order_max + 1, 2):
        stop = 2 * j + 1 + counter
        diag_lb[counter:stop] = (j * (j + 1)) ** 2
        counter = stop

    return np.diag(diag_lb)



import numpy as np
from scipy.ndimage import affine_transform

from dipy.align.reslice import reslice
from dipy.testing.decorators import warning_for_keywords


@warning_for_keywords()
def normalize(image, *, min_v=None, max_v=None, new_min=-1, new_max=1):
    """
    normalization function

    Parameters
    ----------
    image : np.ndarray
        Image to be normalized.
    min_v : int or float, optional
        minimum value range for normalization
        intensities below min_v will be clipped
        if None it is set to min value of image
        Default : None
    max_v : int or float, optional
        maximum value range for normalization
        intensities above max_v will be clipped
        if None it is set to max value of image
        Default : None
    new_min : int or float, optional
        new minimum value after normalization
        Default : 0
    new_max : int or float, optional
        new maximum value after normalization
        Default : 1

    Returns
    -------
    np.ndarray
        Normalized image from range new_min to new_max
    """
    if min_v is None:
        min_v = np.min(image)
    if max_v is None:
        max_v = np.max(image)
    return np.interp(image, (min_v, max_v), (new_min, new_max))


def unnormalize(image, norm_min, norm_max, min_v, max_v):
    """
    unnormalization function

    Parameters
    ----------
    image : np.ndarray
    norm_min : int or float
        minimum value of normalized image
    norm_max : int or float
        maximum value of normalized image
    min_v : int or float
        minimum value of unnormalized image
    max_v : int or float
        maximum value of unnormalized image

    Returns
    -------
    np.ndarray
        unnormalized image from range min_v to max_v
    """
    return (image - norm_min) / (norm_max - norm_min) * (max_v - min_v) + min_v


def set_logger_level(log_level, logger):
    """Change the logger to one of the following:
    DEBUG, INFO, WARNING, CRITICAL, ERROR

    Parameters
    ----------
    log_level : str
        Log level for the logger
    """
    logger.setLevel(level=log_level)


@warning_for_keywords()
def transform_img(
    image,
    affine,
    *,
    voxsize=None,
    considered_points="corners",
    init_shape=(256, 256, 256),
    scale=2,
):
    """
    Function to reshape image as an input to the model

    Parameters
    ----------
    image : np.ndarray
        Image to transform to voxelspace
    affine : np.ndarray
        Affine matrix provided by the file
    voxsize : np.ndarray (3,), optional
        Voxel size of the image
    considered_points : str, optional
        which points to consider when calculating
        the boundary of the image. If there is shearing
        in the affine, 'all' might be more accurate
    init_shape : list, tuple or numpy array (3,), optional
        Initial shape to transform the image to
    scale : float, optional
        How much we want to scale the image

    Returns
    -------
    tuple
        Tuple with variables for recover_img
    """
    if voxsize is not None and np.any(voxsize != np.ones(3)):
        image2, affine2 = reslice(image, affine, voxsize, (1, 1, 1))
    else:
        image2 = image.copy()
        affine2 = affine.copy()
    shape = image2.shape

    if considered_points == "corners":
        corners = np.array(
            [
                [0, 0, 0, 1],
                [shape[0] - 1, 0, 0, 1],
                [0, shape[1] - 1, 0, 1],
                [0, 0, shape[2] - 1, 1],
                [shape[0] - 1, shape[1] - 1, shape[2] - 1, 1],
                [shape[0] - 1, 0, shape[2] - 1, 1],
                [0, shape[1] - 1, shape[2] - 1, 1],
                [shape[0] - 1, shape[1] - 1, 0, 1],
            ],
            dtype=np.float64,
        )
    else:
        temp1 = np.arange(shape[0])
        temp2 = np.arange(shape[1])
        temp3 = np.arange(shape[2])
        grid1, grid2, grid3 = np.meshgrid(temp1, temp2, temp3)
        corners = np.vstack([grid1.ravel(), grid2.ravel(), grid3.ravel()]).T
        corners = np.hstack([corners, np.full((corners.shape[0], 1), 1)])
        corners = corners.astype(np.float64)

    transformed_corners = (affine2 @ corners.T).T
    min_bounds = transformed_corners.min(axis=0)[:3]
    max_bounds = transformed_corners.max(axis=0)[:3]

    # Calculate the required offset to ensure
    # all necessary coordinates are positive
    offset = np.floor(-min_bounds)
    new_shape = (np.ceil(max_bounds) + offset).astype(int)
    offset_array = np.array(
        [[1, 0, 0, offset[0]], [0, 1, 0, offset[1]], [0, 0, 1, offset[2]], [0, 0, 0, 1]]
    )

    new_affine = affine2.copy()
    new_affine = np.matmul(offset_array, new_affine)

    inv_affine = np.linalg.inv(new_affine)
    new_image = np.zeros(tuple(new_shape))
    affine_transform(
        image2, inv_affine, output_shape=tuple(new_shape), output=new_image
    )

    new_image, pad_vs, crop_vs = pad_crop(new_image, init_shape)

    if scale != 1:
        new_image, _ = reslice(new_image, np.eye(4), (1, 1, 1), (scale, scale, scale))

    return new_image, inv_affine, image2.shape, offset_array, scale, crop_vs, pad_vs


def recover_img(
    image,
    inv_affine,
    mid_shape,
    ori_shape,
    offset_array,
    voxsize,
    scale,
    crop_vs,
    pad_vs,
):
    """
    Function to recover image from transform_img

    Parameters
    ----------
    image : np.ndarray
        Image to recover
    inv_affine : np.ndarray
        Affine matrix returned from transform_img
    mid_shape : np.ndarray (3,)
        shape of image returned from transform_img
    ori_shape : tuple (3,)
        original shape of the image
    offset_array : np.ndarray
        Affine matrix that was used in transform_img
        to translate the center
    voxsize : np.ndarray (3,)
        Voxel size used in transform_img
    scale : float
        Scale used in transform_img
    crop_vs : np.ndarray (3,2)
        crop range used in transform_img
    pad_vs : np.ndarray (3,2)
        pad range used in transform_img

    Returns
    -------
    image2 : np.ndarray
        Recovered image
    """
    new_affine = np.linalg.inv(inv_affine)
    new_image, _ = reslice(image, np.eye(4), (scale, scale, scale), (1, 1, 1))
    crop_vs = crop_vs.astype(int)
    pad_vs = pad_vs.astype(int)
    new_image = np.pad(
        new_image,
        (
            (crop_vs[0, 0], crop_vs[0, 1]),
            (crop_vs[1, 0], crop_vs[1, 1]),
            (crop_vs[2, 0], crop_vs[2, 1]),
        ),
    )
    new_image = new_image[
        pad_vs[0, 0] : new_image.shape[0] - pad_vs[0, 1],
        pad_vs[1, 0] : new_image.shape[1] - pad_vs[1, 1],
        pad_vs[2, 0] : new_image.shape[2] - pad_vs[2, 1],
    ]
    new_image = affine_transform(new_image, new_affine, output_shape=mid_shape)
    affine = np.matmul(np.linalg.inv(offset_array), new_affine)
    if voxsize is not None and np.any(voxsize != np.ones(3)):
        image2, _ = reslice(new_image, affine, (1, 1, 1), voxsize)
    else:
        image2 = new_image

    # because of zoom rounding errors
    image2, _, _ = pad_crop(image2, ori_shape)
    return image2


def pad_crop(image, target_shape):
    """
    Function to figure out pad and crop range
    to fit the target shape with the image

    Parameters
    ----------
    image : np.ndarray
        Target image
    target_shape : (3,)
        Target shape

    Returns
    -------
    image : np.ndarray
        Padded/cropped image
    pad_vs : np.ndarray (3,2)
        Pad range used
    crop_vs : np.ndarray (3,2)
        Crop range used
    """
    crop_vs = np.zeros((3, 2)).astype(int)
    pad_vs = np.zeros((3, 2)).astype(int)
    if image.shape == target_shape:
        return image, pad_vs, crop_vs

    pad_crop_v = np.array(target_shape) - np.array(image.shape)
    for d in range(3):
        if pad_crop_v[d] < 0:
            crop_vs[d, 0] = -pad_crop_v[d] // 2
            crop_vs[d, 1] = -pad_crop_v[d] - crop_vs[d, 0]
        elif pad_crop_v[d] > 0:
            pad_vs[d, 0] = pad_crop_v[d] // 2
            pad_vs[d, 1] = pad_crop_v[d] - pad_vs[d, 0]
    image = np.pad(
        image,
        (
            (pad_vs[0, 0], pad_vs[0, 1]),
            (pad_vs[1, 0], pad_vs[1, 1]),
            (pad_vs[2, 0], pad_vs[2, 1]),
        ),
    )
    image = image[
        crop_vs[0, 0] : image.shape[0] - crop_vs[0, 1],
        crop_vs[1, 0] : image.shape[1] - crop_vs[1, 1],
        crop_vs[2, 0] : image.shape[2] - crop_vs[2, 1],
    ]

    return image, pad_vs, crop_vs



"""Utility functions for checking different stuffs."""

import operator

from packaging import version


def check_version(package_name, ver, *, operator=operator.ge):
    """Check if the package is installed and the version satisfies the operator.

    Parameters
    ----------
    package_name : str
        The name of the package to check.
    ver : str
        The version to check against.
    operator : callable, optional
        The operator to use for the comparison.

    Returns
    -------
    bool
        True if the package is installed and the version satisfies the operator.
    """
    try:
        pkg = __import__(package_name)
    except ImportError:
        return False
    if hasattr(pkg, "__version__"):
        return operator(version.parse(pkg.__version__), version.parse(ver))
    return False


def check_min_version(package_name, min_version, *, strict=False):
    """Check if the package is installed and the version is at least min_version.

    Parameters
    ----------
    package_name : str
        The name of the package to check.
    min_version : str
        The minimum version required.
    strict : bool, optional
        If True, the version must be strictly greater than min_version.

    Returns
    -------
    bool
        True if the package is installed and the version is at least min_version.
    """
    op = operator.gt if strict else operator.ge
    return check_version(package_name, min_version, operator=op)


def check_max_version(package_name, max_version, *, strict=False):
    """Check if the package is installed and the version is at most max_version.

    Parameters
    ----------
    package_name : str
        The name of the package to check.
    max_version : str
        The maximum version required.
    strict : bool, optional
        If True, the version must be strictly less than max_version.

    Returns
    -------
    bool
        True if the package is installed and the version is at most max_version.
    """
    op = operator.lt if strict else operator.le
    return check_version(package_name, max_version, operator=op)



"""
===============================================================================
Suppress Gibbs oscillations
===============================================================================

Magnetic Resonance (MR) images are reconstructed from the Fourier coefficients
of acquired k-space images. Since only a finite number of Fourier coefficients
can be acquired in practice, reconstructed MR images can be corrupted by Gibbs
artefacts, which is manifested by intensity oscillations adjacent to edges of
different tissues types :footcite:p:`Veraart2016a`. Although this artefact
affects MR images in general, in the context of diffusion-weighted imaging,
Gibbs oscillations can be magnified in derived diffusion-based estimates
:footcite:p:`Veraart2016a`, :footcite:p:`Perrone2015`.

In the following example, we show how to suppress Gibbs artefacts of MR images.
This algorithm is based on an adapted version of a sub-voxel Gibbs suppression
procedure :footcite:p:`Kellner2016`. Full details of the implemented algorithm
can be found in chapter 3 of :footcite:p:`NetoHenriques2018` (please cite
:footcite:p:`Kellner2016`, :footcite:p:`NetoHenriques2018` if you are using this
code).

The algorithm to suppress Gibbs oscillations can be imported from the denoise
module of dipy:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.data import get_fnames, read_cenir_multib
from dipy.denoise.gibbs import gibbs_removal
from dipy.io.image import load_nifti_data
import dipy.reconst.msdki as msdki
from dipy.segment.mask import median_otsu

###############################################################################
# We first apply this algorithm to a T1-weighted dataset which can be fetched
# using the following code:

t1_fname, t1_denoised_fname, ap_fname = get_fnames(name="tissue_data")
t1 = load_nifti_data(t1_denoised_fname)

###############################################################################
# Let's plot a slice of this dataset.
axial_slice = 88
t1_slice = t1[..., axial_slice]

fig = plt.figure(figsize=(15, 4))
fig.subplots_adjust(wspace=0.2)

t1_slice = np.rot90(t1_slice)

plt.subplot(1, 2, 1)
plt.imshow(t1_slice, cmap="gray", vmin=100, vmax=400)
plt.colorbar()
fig.savefig("structural.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Representative slice of a T1-weighted structural image.
#
#
# Due to the high quality of the data, Gibbs artefacts are not visually
# evident in this dataset. Therefore, to analyse the benefits of the Gibbs
# suppression algorithm, Gibbs artefacts are artificially introduced by
# removing high frequencies of the image's Fourier transform.

c = np.fft.fft2(t1_slice)
c = np.fft.fftshift(c)
N = c.shape[0]
c_crop = c[64:192, 64:192]
N = c_crop.shape[0]
t1_gibbs = abs(np.fft.ifft2(c_crop) / 4)

###############################################################################
# Gibbs oscillation suppression of this single data slice can be performed by
# running the following command:

t1_unring = gibbs_removal(t1_gibbs, inplace=False)

###############################################################################
# Let’s plot the results:

fig1, ax = plt.subplots(1, 3, figsize=(12, 6), subplot_kw={"xticks": [], "yticks": []})

ax.flat[0].imshow(t1_gibbs, cmap="gray", vmin=100, vmax=400)
ax.flat[0].annotate(
    "Rings",
    fontsize=10,
    xy=(81, 70),
    color="red",
    xycoords="data",
    xytext=(30, 0),
    textcoords="offset points",
    arrowprops={"arrowstyle": "->", "color": "red"},
)
ax.flat[0].annotate(
    "Rings",
    fontsize=10,
    xy=(75, 50),
    color="red",
    xycoords="data",
    xytext=(30, 0),
    textcoords="offset points",
    arrowprops={"arrowstyle": "->", "color": "red"},
)

ax.flat[1].imshow(t1_unring, cmap="gray", vmin=100, vmax=400)
ax.flat[1].annotate(
    "WM/GM",
    fontsize=10,
    xy=(75, 50),
    color="green",
    xycoords="data",
    xytext=(30, 0),
    textcoords="offset points",
    arrowprops={"arrowstyle": "->", "color": "green"},
)

ax.flat[2].imshow(t1_unring - t1_gibbs, cmap="gray", vmin=0, vmax=10)
ax.flat[2].annotate(
    "Rings",
    fontsize=10,
    xy=(81, 70),
    color="red",
    xycoords="data",
    xytext=(30, 0),
    textcoords="offset points",
    arrowprops={"arrowstyle": "->", "color": "red"},
)
ax.flat[2].annotate(
    "Rings",
    fontsize=10,
    xy=(75, 50),
    color="red",
    xycoords="data",
    xytext=(30, 0),
    textcoords="offset points",
    arrowprops={"arrowstyle": "->", "color": "red"},
)
plt.show()
fig1.savefig("Gibbs_suppression_structural.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Uncorrected and corrected structural images are shown in the left
# and middle panels, while the difference between these images is shown
# in the right panel.
#
#
# The image artificially corrupted with Gibb's artefacts is shown in the left
# panel. In this panel, the characteristic ringing profile of Gibbs artefacts
# can be visually appreciated (see intensity oscillations pointed by the red
# arrows). The corrected image is shown in the middle panel. One can appreciate
# that artefactual oscillations are visually suppressed without compromising
# the contrast between white and grey matter (e.g. details pointed by the green
# arrow). The difference between uncorrected and corrected data is plotted in
# the right panel which highlights the suppressed Gibbs ringing profile.
#
#
# Now let's show how to use the Gibbs suppression algorithm in
# diffusion-weighted images. We fetch the multi-shell diffusion-weighted
# dataset which was kindly supplied by Romain Valabrègue,
# CENIR, ICM, Paris :footcite:p:`Valabregue2015`.


bvals = [200, 400, 1000, 2000]

img, gtab = read_cenir_multib(bvals=bvals)

data = np.asarray(img.dataobj)

###############################################################################
# For illustration purposes, we select two slices of this dataset

data_slices = data[:, :, 40:42, :]

###############################################################################
# Gibbs oscillation suppression of all multi-shell data and all slices
# can be performed in the following way:

data_corrected = gibbs_removal(data_slices, slice_axis=2, num_processes=-1)

###############################################################################
# Due to the high dimensionality of diffusion-weighted data, we recommend
# that you specify which is the axis of data matrix that corresponds to
# different slices in the above step. This is done by using the optional
# parameter 'slice_axis'.
#
# Below we plot the results for an image acquired with b-value=0:

fig2, ax = plt.subplots(1, 3, figsize=(12, 6), subplot_kw={"xticks": [], "yticks": []})

ax.flat[0].imshow(
    data_slices[:, :, 0, 0].T, cmap="gray", origin="lower", vmin=0, vmax=10000
)
ax.flat[0].set_title("Uncorrected b0")
ax.flat[1].imshow(
    data_corrected[:, :, 0, 0].T, cmap="gray", origin="lower", vmin=0, vmax=10000
)
ax.flat[1].set_title("Corrected b0")
ax.flat[2].imshow(
    data_corrected[:, :, 0, 0].T - data_slices[:, :, 0, 0].T,
    cmap="gray",
    origin="lower",
    vmin=-500,
    vmax=500,
)
ax.flat[2].set_title("Gibbs residuals")

plt.show()
fig2.savefig("Gibbs_suppression_b0.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Uncorrected (left panel) and corrected (middle panel) b-value=0 images. For
# reference, the difference between uncorrected and corrected images is shown
# in the right panel.
#
#
# The above figure shows that the benefits of suppressing Gibbs artefacts is
# hard to observe on b-value=0 data. Therefore, diffusion derived metrics for
# both uncorrected and corrected data are computed using the mean signal
# diffusion kurtosis image technique
# (:ref:`sphx_glr_examples_built_reconstruction_reconst_msdki.py`).
#
# To avoid unnecessary calculations on the background of the image, we also
# compute a brain mask.

# Create a brain mask
maskdata, mask = median_otsu(
    data_slices,
    vol_idx=range(10, 50),
    median_radius=3,
    numpass=1,
    autocrop=False,
    dilate=1,
)

# Define mean signal diffusion kurtosis model
dki_model = msdki.MeanDiffusionKurtosisModel(gtab)

# Fit the uncorrected data
dki_fit = dki_model.fit(data_slices, mask=mask)
MSKini = dki_fit.msk

# Fit the corrected data
dki_fit = dki_model.fit(data_corrected, mask=mask)
MSKgib = dki_fit.msk

###############################################################################
# Let's plot the results

fig3, ax = plt.subplots(1, 3, figsize=(12, 12), subplot_kw={"xticks": [], "yticks": []})

ax.flat[0].imshow(MSKini[:, :, 0].T, cmap="gray", origin="lower", vmin=0, vmax=1.5)
ax.flat[0].set_title("MSK (uncorrected)")
ax.flat[0].annotate(
    "Rings",
    fontsize=12,
    xy=(59, 63),
    color="red",
    xycoords="data",
    xytext=(45, 0),
    textcoords="offset points",
    arrowprops={"arrowstyle": "->", "color": "red"},
)

ax.flat[1].imshow(MSKgib[:, :, 0].T, cmap="gray", origin="lower", vmin=0, vmax=1.5)
ax.flat[1].set_title("MSK (corrected)")

ax.flat[2].imshow(
    MSKgib[:, :, 0].T - MSKini[:, :, 0].T,
    cmap="gray",
    origin="lower",
    vmin=-0.2,
    vmax=0.2,
)
ax.flat[2].set_title("MSK (uncorrected - corrected")
ax.flat[2].annotate(
    "Rings",
    fontsize=12,
    xy=(59, 63),
    color="red",
    xycoords="data",
    xytext=(45, 0),
    textcoords="offset points",
    arrowprops={"arrowstyle": "->", "color": "red"},
)

plt.show()
fig3.savefig("Gibbs_suppression_msdki.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Uncorrected and corrected mean signal kurtosis images are shown in the left
# and middle panels. The difference between uncorrected and corrected images
# are show in the right panel.
#
#
# In the left panel of the figure above, Gibbs artefacts can be appreciated by
# the negative values of mean signal kurtosis (black voxels) adjacent to the
# brain ventricle (red arrows). These negative values seem to be suppressed
# after the `gibbs_removal` function is applied. For a better visualization of
# Gibbs oscillations, the difference between corrected and uncorrected images
# are shown in the right panel.
#
#
# References
# ----------
#
# .. footbibliography::
#



#!/usr/bin/python
"""Class and helper functions for fitting the Histological ResDNN model."""

import logging

import numpy as np

from dipy.core.gradients import get_bval_indices, unique_bvals_magnitude
from dipy.core.sphere import HemiSphere
from dipy.data import get_fnames, get_sphere
from dipy.nn.utils import set_logger_level
from dipy.reconst.shm import sf_to_sh, sh_to_sf, sph_harm_ind_list
from dipy.testing.decorators import doctest_skip_parser
from dipy.utils.optpkg import optional_package

torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
if have_torch:
    from torch.nn import Linear, Module
else:

    class Module:
        pass

    logging.warning(
        "This model requires Pytorch.\
                    Please install these packages using \
                    pip."
    )

logging.basicConfig()
logger = logging.getLogger("histo_resdnn")


class DenseModel(Module):
    def __init__(self, sh_size, num_hidden):
        super(DenseModel, self).__init__()
        self.fc1 = Linear(sh_size, 400)
        self.fc2 = Linear(400, num_hidden)
        self.fc3 = Linear(num_hidden, 200)
        self.fc4 = Linear(200, num_hidden)
        self.fc5 = Linear(num_hidden, 200)
        self.fc6 = Linear(200, num_hidden)

    def forward(self, x):
        x1 = torch.relu(self.fc1(x))
        x2 = torch.relu(self.fc2(x1))
        x3 = torch.relu(self.fc3(x2))
        x4 = self.fc4(x3)

        # Adding x2 and x4
        res_add = x2 + x4

        x5 = torch.relu(self.fc5(res_add))
        x6 = self.fc6(x5)

        return x6


class HistoResDNN:
    """This class is intended for the ResDNN Histology Network model.

    ResDNN :footcite:p:`Nath2019` is a deep neural network that employs residual
    blocks deep neural network to predict ground truth SH coefficients from SH
    coefficients computed using DWI data. To this end, authors considered
    histology FOD-computed SH coefficients (obtained from ex vivo non-human
    primate acquisitions) as their ground truth, and the DWI-computed SH
    coefficients as their target.

    References
    ----------
    .. footbibliography::
    """

    @doctest_skip_parser
    def __init__(
        self, *, sh_order_max=8, basis_type="tournier07", verbose=False, use_cuda=False
    ):
        """Model initialization

        The model was re-trained for usage with a different basis function
        ('tournier07') like the proposed model in :footcite:p:`Nath2019`.

        To obtain the pre-trained model, use::
        >>> resdnn_model = HistoResDNN() # skip if not have_torch
        >>> fetch_model_weights_path = get_fnames(name='histo_resdnn_torch_weights') # skip if not have_torch
        >>> resdnn_model.load_model_weights(fetch_model_weights_path) # skip if not have_torch

        This model is designed to take as input raw DWI signal on a sphere
        (ODF) represented as SH of order 8 in the tournier basis and predict
        fODF of order 8 in the tournier basis. Effectively, this model is
        mimicking a CSD fit.

        Parameters
        ----------
        sh_order_max : int, optional
            Maximum SH order (l) in the SH fit.  For ``sh_order_max``, there
            will be
            ``(sh_order_max + 1) * (sh_order_max + 2) / 2`` SH coefficients
            for a symmetric basis.
        basis_type : {'tournier07', 'descoteaux07'}, optional
            ``tournier07`` (default) or ``descoteaux07``.
        verbose : bool, optional
            Whether to show information about the processing.
        use_cuda : bool, optional
            Whether to use GPU for processing.
            If False or no CUDA is detected, CPU will be used.

        References
        ----------
        .. footbibliography::
        """  # noqa: E501
        if not have_torch:
            raise torch()

        self.sh_order_max = sh_order_max
        self.sh_size = len(sph_harm_ind_list(sh_order_max)[0])
        self.basis_type = basis_type

        log_level = "INFO" if verbose else "CRITICAL"
        set_logger_level(log_level, logger)
        if self.basis_type != "tournier07":
            logger.warning(
                "Be careful, original weights were obtained "
                "from training on the tournier07 basis, "
                "unless you re-trained the network, do not "
                "change basis!"
            )

        # ResDNN Network Flow
        num_hidden = self.sh_size
        self.model = DenseModel(self.sh_size, num_hidden).type(torch.float64)
        if use_cuda and torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")
        self.model = self.model.to(self.device)

    def fetch_default_weights(self):
        """Load the model pre-training weights to use for the fitting.
        Will not work if the declared SH_ORDER does not match the weights
        expected input.
        """
        fetch_model_weights_path = get_fnames(name="histo_resdnn_torch_weights")
        self.load_model_weights(fetch_model_weights_path)

    def load_model_weights(self, weights_path):
        """Load the custom pre-training weights to use for the fitting.
        Will not work if the declared SH_ORDER does not match the weights
        expected input.

        The weights for a sh_order of 8 can be obtained via the function:
            get_fnames('histo_resdnn_torch_weights').

        Parameters
        ----------
        weights_path : str
            Path to the file containing the weights (pth, saved by Pytorch)
        """
        try:
            self.model.load_state_dict(
                torch.load(
                    weights_path,
                    weights_only=True,
                    map_location=self.device,
                )
            )
            self.model.eval()
        except RuntimeError as e:
            raise RuntimeError(
                "Expected input for the provided model weights do not match the "
                f"declared model ({self.sh_size})"
            ) from e

    def __predict(self, x_test):
        """Predict fODF (as SH) from input raw DWI signal (as SH)

        Parameters
        ----------
        x_test : np.ndarray
            Array of size (N, M) where M is
            ``(sh_order_max + 1) * (sh_order_max + 2) / 2``.
            N should not be too big as to limit memory usage.

        Returns
        -------
        np.ndarray (N, M)
            Predicted fODF (as SH)
        """
        if x_test.shape[-1] != self.sh_size:
            raise ValueError(
                "Expected input for the provided model weights do not match the "
                f"declared model ({self.sh_size})"
            )

        return self.model(torch.from_numpy(x_test)).detach().numpy()

    def predict(self, data, gtab, *, mask=None, chunk_size=1000):
        """Wrapper function to facilitate prediction of larger dataset.
        The function will mask, normalize, split, predict and 're-assemble'
        the data as a volume.

        Parameters
        ----------
        data : np.ndarray
            DWI signal in a 4D array
        gtab : GradientTable class instance
            The acquisition scheme matching the data (must contain at least
            one b0)
        mask : np.ndarray, optional
            Binary mask of the brain to avoid unnecessary computation and
            unreliable prediction outside the brain.
            Default: Compute prediction only for nonzero voxels (with at least
            one nonzero DWI value).
        chunk_size : int, optional
            Batch size when running model prediction.

        Returns
        -------
        pred_sh_coef : np.ndarray (x, y, z, M)
            Predicted fODF (as SH). The volume has matching shape to the input
            data, but with
            ``(sh_order_max + 1) * (sh_order_max + 2) / 2`` as a last
            dimension.
        """
        if mask is None:
            logger.warning(
                "Mask should be provided to accelerate "
                "computation, and because predictions are "
                "not reliable outside of the brain."
            )
            mask = np.sum(data, axis=-1)
        mask = mask.astype(bool)

        # Extract B0's and obtain a mean B0
        b0_indices = gtab.b0s_mask
        if not len(b0_indices) > 0:
            raise ValueError("b0 must be present for DWI normalization.")
        logger.info(f"b0 indices found are: {np.argwhere(b0_indices).ravel()}")

        mean_b0 = np.mean(data[..., b0_indices], axis=-1)

        # Detect number of b-values and extract a single shell of DW-MRI Data
        unique_shells = np.sort(unique_bvals_magnitude(gtab.bvals))
        logger.info(f"Number of b-values: {unique_shells}")

        # Extract DWI only
        dw_indices = get_bval_indices(gtab.bvals, unique_shells[1])
        dw_data = data[..., dw_indices]
        dw_bvecs = gtab.bvecs[dw_indices, :]

        # Normalize the DW-MRI Data with the mean b0 (voxel-wise)
        norm_dw_data = np.zeros(dw_data.shape)
        for n in range(len(dw_indices)):
            norm_dw_data[..., n] = np.divide(
                dw_data[..., n], mean_b0, where=np.abs(mean_b0) > 0.000001
            )

        # Fit SH to the raw DWI signal
        h_sphere = HemiSphere(xyz=dw_bvecs)
        dw_sh_coef = sf_to_sh(
            norm_dw_data,
            h_sphere,
            smooth=0.0006,
            basis_type=self.basis_type,
            sh_order_max=self.sh_order_max,
        )

        # Flatten and mask the data (N, SH_SIZE) to facilitate chunks
        ori_shape = dw_sh_coef.shape
        flat_dw_sh_coef = dw_sh_coef[mask > 0]
        flat_pred_sh_coef = np.zeros(flat_dw_sh_coef.shape)

        count = len(flat_dw_sh_coef) // chunk_size
        for i in range(count + 1):
            if i % 100 == 0 or i == count:
                logger.info(f"Chunk #{i} out of {count}")
            tmp_sh = self.__predict(
                flat_dw_sh_coef[i * chunk_size : (i + 1) * chunk_size]
            )

            # Removing negative values from the SF
            sphere = get_sphere(name="repulsion724")
            tmp_sf = sh_to_sf(
                sh=tmp_sh,
                sphere=sphere,
                basis_type=self.basis_type,
                sh_order_max=self.sh_order_max,
            )
            tmp_sf[tmp_sf < 0] = 0
            tmp_sh = sf_to_sh(
                tmp_sf,
                sphere,
                smooth=0.0006,
                basis_type=self.basis_type,
                sh_order_max=self.sh_order_max,
            )
            flat_pred_sh_coef[i * chunk_size : (i + 1) * chunk_size] = tmp_sh

        pred_sh_coef = np.zeros(ori_shape)
        pred_sh_coef[mask > 0] = flat_pred_sh_coef

        return pred_sh_coef



#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True

import numpy as np
cimport numpy as cnp

cimport safe_openmp as openmp
from safe_openmp cimport have_openmp

from cython.parallel import prange
from libc.stdlib cimport malloc, free
from libc.math cimport sqrt

from dipy.utils.omp import determine_num_threads
from dipy.utils.omp cimport set_num_threads, restore_default_num_threads

cdef cnp.dtype f64_dt = np.dtype(np.float64)


cdef double min_direct_flip_dist(double *a,double *b,
                                 cnp.npy_intp rows) noexcept nogil:
    r""" Minimum of direct and flip average (MDF) distance between two
    streamlines.

    See :footcite:p:`Garyfallidis2012a` for a definition of the distance.

    Parameters
    ----------
    a : double pointer
        first streamline
    b : double pointer
        second streamline
    rows : number of points of the streamline
        both tracks need to have the same number of points

    Returns
    -------
    out : double
        minimum of direct and flipped average distances

    References
    ----------
    .. footbibliography::
    """

    cdef:
        cnp.npy_intp i=0, j=0
        double sub=0, subf=0, distf=0, dist=0, tmprow=0, tmprowf=0


    for i in range(rows):
        tmprow = 0
        tmprowf = 0
        for j in range(3):
            sub = a[i * 3 + j] - b[i * 3 + j]
            subf = a[i * 3 + j] - b[(rows - 1 - i) * 3 + j]
            tmprow += sub * sub
            tmprowf += subf * subf
        dist += sqrt(tmprow)
        distf += sqrt(tmprowf)

    dist = dist / <double>rows
    distf = distf / <double>rows

    if dist <= distf:
        return dist
    return distf


def _bundle_minimum_distance_matrix(double [:, ::1] static,
                                    double [:, ::1] moving,
                                    cnp.npy_intp static_size,
                                    cnp.npy_intp moving_size,
                                    cnp.npy_intp rows,
                                    double [:, ::1] D,
                                    num_threads=None):
    """ MDF-based pairwise distance optimization function

    We minimize the distance between moving streamlines of the same number of
    points as they align with the static streamlines.

    Parameters
    ----------
    static: array
        Static streamlines
    moving: array
        Moving streamlines
    static_size : int
        Number of static streamlines
    moving_size : int
        Number of moving streamlines
    rows : int
        Number of points per streamline
    D : 2D array
        Distance matrix
    num_threads : int, optional
        Number of threads to be used for OpenMP parallelization. If None
        (default) the value of OMP_NUM_THREADS environment variable is used
        if it is set, otherwise all available threads are used. If < 0 the
        maximal number of threads minus |num_threads + 1| is used (enter -1 to
        use as many threads as possible). 0 raises an error.

    Returns
    -------
    cost : double
    """

    cdef:
        cnp.npy_intp i=0, j=0, mov_i=0, mov_j=0
        int threads_to_use = -1

    threads_to_use = determine_num_threads(num_threads)
    set_num_threads(threads_to_use)

    with nogil:

        for i in prange(static_size):
            for j in prange(moving_size):

                D[i, j] = min_direct_flip_dist(&static[i * rows, 0],
                                               &moving[j * rows, 0],
                                               rows)

    if num_threads is not None:
        restore_default_num_threads()

    return np.asarray(D)


def _bundle_minimum_distance(double [:, ::1] static,
                             double [:, ::1] moving,
                             cnp.npy_intp static_size,
                             cnp.npy_intp moving_size,
                             cnp.npy_intp rows,
                             num_threads=None):
    """ MDF-based pairwise distance optimization function

    We minimize the distance between moving streamlines of the same number of
    points as they align with the static streamlines.

    Parameters
    ----------
    static : array
        Static streamlines
    moving : array
        Moving streamlines
    static_size : int
        Number of static streamlines
    moving_size : int
        Number of moving streamlines
    rows : int
        Number of points per streamline
    num_threads : int, optional
        Number of threads to be used for OpenMP parallelization. If None
        (default) the value of OMP_NUM_THREADS environment variable is used
        if it is set, otherwise all available threads are used. If < 0 the
        maximal number of threads minus |num_threads + 1| is used (enter -1 to
        use as many threads as possible). 0 raises an error.

    Returns
    -------
    cost : double

    Notes
    -----
    The difference with ``_bundle_minimum_distance_matrix`` is that it does not
    save the full distance matrix and therefore needs much less memory.
    """

    cdef:
        cnp.npy_intp i=0, j=0
        double sum_i=0, sum_j=0, tmp=0
        double inf = np.finfo('f8').max
        double dist=0
        double * min_j
        double * min_i
        openmp.omp_lock_t lock
        int threads_to_use = -1

    threads_to_use = determine_num_threads(num_threads)
    set_num_threads(threads_to_use)

    with nogil:

        if have_openmp:
            openmp.omp_init_lock(&lock)

        min_j = <double *> malloc(static_size * sizeof(double))
        min_i = <double *> malloc(moving_size * sizeof(double))

        for i in range(static_size):
            min_j[i] = inf

        for j in range(moving_size):
            min_i[j] = inf

        for i in prange(static_size):

            for j in range(moving_size):

                tmp = min_direct_flip_dist(&static[i * rows, 0],
                                       &moving[j * rows, 0], rows)

                if have_openmp:
                    openmp.omp_set_lock(&lock)
                if tmp < min_j[i]:
                    min_j[i] = tmp

                if tmp < min_i[j]:
                    min_i[j] = tmp
                if have_openmp:
                    openmp.omp_unset_lock(&lock)

        if have_openmp:
            openmp.omp_destroy_lock(&lock)

        for i in range(static_size):
            sum_i += min_j[i]

        for j in range(moving_size):
            sum_j += min_i[j]

        free(min_j)
        free(min_i)

        dist = (sum_i / <double>static_size + sum_j / <double>moving_size)

        dist = 0.25 * dist * dist

    if num_threads is not None:
        restore_default_num_threads()

    return dist



def _bundle_minimum_distance_asymmetric(double [:, ::1] static,
                                        double [:, ::1] moving,
                                        cnp.npy_intp static_size,
                                        cnp.npy_intp moving_size,
                                        cnp.npy_intp rows):
    """ MDF-based pairwise distance optimization function

    We minimize the distance between moving streamlines of the same number of
    points as they align with the static streamlines.

    Parameters
    ----------
    static : array
        Static streamlines
    moving : array
        Moving streamlines
    static_size : int
        Number of static streamlines
    moving_size : int
        Number of moving streamlines
    rows : int
        Number of points per streamline

    Returns
    -------
    cost : double

    Notes
    -----
    The difference with ``_bundle_minimum_distance`` is that we sum the
    minimum values only for the static. Therefore, this is an asymmetric
    distance metric. This means that we are weighting only one direction of the
    registration. Not both directions. This can be very useful when we want
    to register a big set of bundles to a small set of bundles.
    See :footcite:p:`Wanyan2017`.

    References
    ----------
    .. footbibliography::

    """

    cdef:
        cnp.npy_intp i=0, j=0
        double sum_i=0, sum_j=0, tmp=0
        double inf = np.finfo('f8').max
        double dist=0
        double * min_j
        openmp.omp_lock_t lock

    with nogil:

        if have_openmp:
            openmp.omp_init_lock(&lock)

        min_j = <double *> malloc(static_size * sizeof(double))

        for sz_i in range(static_size):
            min_j[sz_i] = inf

        for i in prange(static_size):

            for j in range(moving_size):

                tmp = min_direct_flip_dist(&static[i * rows, 0],
                                           &moving[j * rows, 0], rows)

                if have_openmp:
                    openmp.omp_set_lock(&lock)
                if tmp < min_j[i]:
                    min_j[i] = tmp

                if have_openmp:
                    openmp.omp_unset_lock(&lock)

        if have_openmp:
            openmp.omp_destroy_lock(&lock)

        for i in range(static_size):
            sum_i += min_j[i]

        free(min_j)

        dist = sum_i / <double>static_size

    return dist


def distance_matrix_mdf(streamlines_a, streamlines_b):
    r""" Minimum direct flipped distance matrix between two streamline sets

    All streamlines need to have the same number of points

    Parameters
    ----------
    streamlines_a : sequence
       of streamlines as arrays, [(N, 3) .. (N, 3)]
    streamlines_b : sequence
       of streamlines as arrays, [(N, 3) .. (N, 3)]

    Returns
    -------
    DM : array, shape (len(streamlines_a), len(streamlines_b))
        distance matrix
    """
    cdef:
        cnp.npy_intp i, j, lentA, lentB
    # preprocess tracks
    cdef:
        cnp.npy_intp longest_track_len = 0, track_len
        longest_track_lenA, longest_track_lenB
        cnp.ndarray[object, ndim=1] tracksA64
        cnp.ndarray[object, ndim=1] tracksB64
        cnp.ndarray[cnp.double_t, ndim=2] DM

    lentA = len(streamlines_a)
    lentB = len(streamlines_b)
    tracksA64 = np.zeros((lentA,), dtype=object)
    tracksB64 = np.zeros((lentB,), dtype=object)
    DM = np.zeros((lentA,lentB), dtype=np.double)
    if streamlines_a[0].shape[0] != streamlines_b[0].shape[0]:
        msg = 'Streamlines should have the same number of points as required'
        msg += 'by the MDF distance'
        raise ValueError(msg)
    # process tracks to predictable memory layout
    for i in range(lentA):
        tracksA64[i] = np.ascontiguousarray(streamlines_a[i], dtype=f64_dt)
    for i in range(lentB):
        tracksB64[i] = np.ascontiguousarray(streamlines_b[i], dtype=f64_dt)
    # preallocate buffer array for track distance calculations
    cdef:
        cnp.float64_t *t1_ptr
        cnp.float64_t *t2_ptr
        cnp.float64_t *min_buffer
    # cycle over tracks
    cdef:
        cnp.ndarray [cnp.float64_t, ndim=2] t1, t2
        cnp.npy_intp t1_len, t2_len
        double d[2]
    t_len = tracksA64[0].shape[0]

    for i from 0 <= i < lentA:
        t1 = tracksA64[i]
        t1_ptr = <cnp.float64_t *> cnp.PyArray_DATA(t1)
        for j from 0 <= j < lentB:
            t2 = tracksB64[j]
            t2_ptr = <cnp.float64_t *> cnp.PyArray_DATA(t2)

            DM[i, j] = min_direct_flip_dist(t1_ptr, t2_ptr,t_len)

    return DM
